import datetime
from collections import OrderedDict

import numpy as np
import pandas as pd
import pytest

from plateau.core.common_metadata import make_meta, read_schema_metadata
from plateau.core.dataset import DatasetMetadata
from plateau.core.uuid import gen_uuid
from plateau.io.eager import (
    create_empty_dataset_header,
    store_dataframes_as_dataset,
    write_single_partition,
)
from plateau.io.testing.write import *  # noqa: F403, F4
from plateau.io_components.metapartition import MetaPartition


def _store_dataframes(dfs, **kwargs):
    # Positional arguments in function but `None` is acceptable input
    for kw in ("dataset_uuid", "store"):
        if kw not in kwargs:
            kwargs[kw] = None

    return store_dataframes_as_dataset(dfs=dfs, **kwargs)


@pytest.fixture()
def bound_store_dataframes():
    return _store_dataframes


def test_write_single_partition(store_factory, mock_uuid, metadata_version):
    create_empty_dataset_header(
        store=store_factory(),
        schema=pd.DataFrame({"col": [1]}),
        dataset_uuid="some_dataset",
        metadata_version=metadata_version,
        table_name="table1",
    )

    new_data = pd.DataFrame({"col": [1, 2]})
    keys_in_store = set(store_factory().keys())
    new_mp = write_single_partition(
        store=store_factory,
        dataset_uuid="some_dataset",
        data=new_data,
        table_name="table1",
    )

    keys_in_store.add("some_dataset/table1/auto_dataset_uuid.parquet")
    assert set(store_factory().keys()) == keys_in_store
    expected_mp = MetaPartition(
        label="auto_dataset_uuid",  # this will be a hash of the input
        file="some_dataset/table1/auto_dataset_uuid.parquet",
        metadata_version=4,
        schema=make_meta(pd.DataFrame({"col": [1, 2]}), origin="table1"),
    )

    assert new_mp == expected_mp

    with pytest.raises(ValueError):
        # col is an integer column so this is incompatible.
        new_data = pd.DataFrame({"col": [datetime.date(2010, 1, 1)]})
        write_single_partition(
            store=store_factory,
            dataset_uuid="some_dataset",
            data=new_data,
            table_name="table1",
        )


def test_create_dataset_header_minimal_version(store, metadata_storage_format):
    with pytest.raises(NotImplementedError):
        create_empty_dataset_header(
            store=store,
            schema=pd.DataFrame({"col": [1]}),
            dataset_uuid="new_dataset_uuid",
            metadata_storage_format=metadata_storage_format,
            metadata_version=3,
        )
    create_empty_dataset_header(
        store=store,
        schema=pd.DataFrame({"col": [1]}),
        dataset_uuid="new_dataset_uuid",
        metadata_storage_format=metadata_storage_format,
        metadata_version=4,
    )


def test_create_dataset_header(store, metadata_storage_format, frozen_time):
    schema = make_meta(pd.DataFrame({"col": [1]}), origin="1")
    new_dataset = create_empty_dataset_header(
        store=store,
        schema=schema,
        dataset_uuid="new_dataset_uuid",
        metadata_storage_format=metadata_storage_format,
        metadata_version=4,
    )

    expected_dataset = DatasetMetadata(
        uuid="new_dataset_uuid",
        metadata_version=4,
        explicit_partitions=False,
        schema=schema,
    )
    assert new_dataset == expected_dataset

    storage_keys = list(store.keys())
    assert len(storage_keys) == 2

    loaded = DatasetMetadata.load_from_store(store=store, uuid="new_dataset_uuid")
    assert loaded == expected_dataset

    # If the read succeeds, the schema is written
    read_schema_metadata(dataset_uuid=new_dataset.uuid, store=store)


# TODO: move `store_dataframes_as_dataset` tests to generic tests or remove if redundant
def test_store_dataframes_as_dataset_no_pipeline_partition_on(store):
    df = pd.DataFrame(
        {"P": np.arange(0, 10), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )

    dataset = store_dataframes_as_dataset(
        store=store,
        dataset_uuid="dataset_uuid",
        dfs=[df],
        partition_on="P",
        metadata_version=4,
    )

    assert isinstance(dataset, DatasetMetadata)
    assert len(dataset.partitions) == 10

    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)

    assert dataset == stored_dataset


def test_store_dataframes_as_dataset_partition_on_inconsistent(store):
    df = pd.DataFrame(
        {"P": np.arange(0, 10), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )

    with pytest.raises(ValueError) as excinfo:
        store_dataframes_as_dataset(
            store=store,
            dataset_uuid="dataset_uuid",
            dfs=[df],
            partition_on=["NOT_IN_COLUMNS"],
            metadata_version=4,
        )
    assert str(excinfo.value) == "Partition column(s) missing: NOT_IN_COLUMNS"


def test_store_dataframes_as_dataset_no_pipeline(metadata_version, store):
    df = pd.DataFrame(
        {"P": np.arange(0, 10), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )

    dataset = store_dataframes_as_dataset(
        store=store,
        dataset_uuid="dataset_uuid",
        dfs=[df],
        metadata_version=metadata_version,
    )

    assert isinstance(dataset, DatasetMetadata)
    assert len(dataset.partitions) == 1
    assert dataset.metadata_version == metadata_version

    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)

    assert dataset == stored_dataset


def test_store_dataframes_as_dataset_mp(metadata_version, store):
    df = pd.DataFrame(
        {"P": np.arange(0, 10), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )

    mp = MetaPartition(
        label=gen_uuid(),
        data=df,
        metadata_version=metadata_version,
    )

    dataset = store_dataframes_as_dataset(
        store=store,
        dataset_uuid="dataset_uuid",
        dfs=[mp],
        metadata_version=metadata_version,
    )

    assert isinstance(dataset, DatasetMetadata)
    assert len(dataset.partitions) == 1
    assert dataset.metadata_version == metadata_version

    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)

    assert dataset == stored_dataset


def test_write_single_partition_different_partitioning(store_factory):
    df = pd.DataFrame(
        OrderedDict([("location", ["0", "1", "2"]), ("other", ["a", "a", "a"])])
    )

    input_ = [df]

    dataset = store_dataframes_as_dataset(
        dfs=input_,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        metadata_version=4,
        partition_on=["other"],
    )

    new_data = [
        pd.DataFrame(
            OrderedDict([("other", ["b", "b", "b"]), ("location", ["0", "1", "2"])])
        )
    ]
    initial_keys = len(list(store_factory().keys()))
    with pytest.raises(ValueError):
        write_single_partition(
            store=store_factory,
            dataset_uuid=dataset.uuid,
            data=new_data,
            partition_on="location",
        )
    assert initial_keys == len(list(store_factory().keys()))
    write_single_partition(
        store=store_factory,
        dataset_uuid=dataset.uuid,
        data=new_data,
        partition_on=["other"],
    )
    assert initial_keys + 1 == len(list(store_factory().keys()))

    # If no partitioning is given, it will be determined based on the existing dataset
    write_single_partition(
        store=store_factory, dataset_uuid=dataset.uuid, data=new_data
    )
    assert initial_keys + 2 == len(list(store_factory().keys()))



import random
from urllib.parse import quote

import minimalkv
import numpy as np
import pandas as pd
import simplejson
from packaging import version

from plateau.core.common_metadata import (
    _get_common_metadata_key,
    make_meta,
    store_schema_metadata,
)
from plateau.core.dataset import DatasetMetadata, create_partition_key, naming
from plateau.core.urlencode import quote_indices, unquote_indices


def test_create_partition_key():
    key = create_partition_key(
        "my-uuid", "testtable", [("index1", "value1"), ("index2", "value2")]
    )
    assert key == "my-uuid/testtable/index1=value1/index2=value2/data"


def test_index_quote_roundtrip():
    indices = [
        (1, b"Muenchen"),
        ("location", b"Muenchen"),
        ("location", "München"),
        ("product", "å\\ øß"),
    ]
    expected = [
        ("1", "Muenchen"),
        ("location", "Muenchen"),
        ("location", "München"),
        ("product", "å\\ øß"),
    ]
    assert expected == unquote_indices(quote_indices(indices))


def testunquote_indices():
    index_strings = [
        "{}={}".format(quote("location"), quote("München".encode())),
        "{}={}".format(quote("product"), quote("å\\ øß".encode())),
    ]
    indices = unquote_indices(index_strings)
    assert indices == [("location", "München"), ("product", "å\\ øß")]


def test_dynamic_partitions(store):
    """Do not specify partitions in metadata, but read them dynamically from
    store."""
    partition_suffix = "suffix"
    dataset_uuid = "uuid+namespace-attribute12_underscored"
    partition0_core = create_partition_key(
        dataset_uuid,
        "core",
        [("location", "L-0")],
        f"{partition_suffix}.parquet",
    )
    partition1_core = create_partition_key(
        dataset_uuid,
        "core",
        [("location", "L-1")],
        f"{partition_suffix}.parquet",
    )
    metadata = {"dataset_metadata_version": 4, "dataset_uuid": dataset_uuid}
    expected_partitions = {
        f"location=L-0/{partition_suffix}": {"files": {"core": partition0_core}},
        f"location=L-1/{partition_suffix}": {"files": {"core": partition1_core}},
    }
    expected_indices = {
        "location": {
            "L-0": [f"location=L-0/{partition_suffix}"],
            "L-1": [f"location=L-1/{partition_suffix}"],
        }
    }

    # put two partitions for two tables each to store
    store.put(
        f"{dataset_uuid}{naming.METADATA_BASE_SUFFIX}.json",
        simplejson.dumps(metadata).encode("utf-8"),
    )
    store.put(partition0_core, b"test")
    store.put(partition1_core, b"test")
    store_schema_metadata(
        make_meta(
            pd.DataFrame({"location": [f"L-0/{partition_suffix}"]}),
            origin="stored",
        ),
        dataset_uuid,
        store,
        "core",
    )

    # instantiate metadata to write table metadatad
    core_schema = make_meta(
        pd.DataFrame(
            {
                "column_0": pd.Series([1], dtype=int),
                "column_1": pd.Series([1], dtype=int),
                "location": pd.Series(["str"]),
            }
        ),
        origin="core",
    )
    store_schema_metadata(core_schema, dataset_uuid, store, "core")

    dmd = DatasetMetadata.load_from_store(dataset_uuid, store)
    dmd = dmd.load_partition_indices()

    dmd_dict = dmd.to_dict()
    assert dmd_dict["partitions"] == expected_partitions
    assert dmd_dict["indices"] == expected_indices


def test_dynamic_partitions_multiple_indices(store):
    """Do not specify partitions in metadata, but read them dynamically from
    store."""
    suffix = "suffix"
    dataset_uuid = "uuid+namespace-attribute12_underscored"
    partition0_core = create_partition_key(
        dataset_uuid,
        "core",
        [("location", "L-0"), ("product", "P-0")],
        f"{suffix}.parquet",
    )
    partition1_core = create_partition_key(
        dataset_uuid,
        "core",
        [("location", "L-1"), ("product", "P-0")],
        f"{suffix}.parquet",
    )
    metadata = {"dataset_metadata_version": 4, "dataset_uuid": dataset_uuid}
    expected_partitions = {
        f"location=L-0/product=P-0/{suffix}": {"files": {"core": partition0_core}},
        f"location=L-1/product=P-0/{suffix}": {"files": {"core": partition1_core}},
    }
    expected_indices = {
        "location": {
            "L-0": [f"location=L-0/product=P-0/{suffix}"],
            "L-1": [f"location=L-1/product=P-0/{suffix}"],
        },
        "product": {
            "P-0": [
                f"location=L-0/product=P-0/{suffix}",
                f"location=L-1/product=P-0/{suffix}",
            ]
        },
    }

    store.put(partition0_core, b"test")
    store.put(partition1_core, b"test")
    store_schema_metadata(
        make_meta(pd.DataFrame({"location": ["L-0"], "product": ["P-0"]}), origin="1"),
        dataset_uuid,
        store,
        "core",
    )

    dmd = DatasetMetadata.load_from_dict(metadata, store)
    dmd = dmd.load_partition_indices()
    dmd_dict = dmd.to_dict()
    assert dmd_dict["partitions"] == expected_partitions
    # Sorting may differ in the index list. This is ok for runtime
    # but does produce flaky tests thus sort them.
    sorted_result = {
        column: {label: sorted(x) for label, x in index.items()}
        for column, index in dmd_dict["indices"].items()
    }
    assert sorted_result == expected_indices


def test_dynamic_partitions_with_garbage(store):
    """In case there are unknown files, dataset and indices still load
    correctly."""
    dataset_uuid = "uuid+namespace-attribute12_underscored"
    partition_suffix = "suffix"
    partition0_core = create_partition_key(
        dataset_uuid,
        "core",
        [("location", "L-0"), ("product", "P-0")],
        f"{partition_suffix}.parquet",
    )
    partition1_core = create_partition_key(
        dataset_uuid,
        "core",
        [("location", "L-1"), ("product", "P-0")],
        f"{partition_suffix}.parquet",
    )
    metadata = {"dataset_metadata_version": 4, "dataset_uuid": dataset_uuid}
    expected_partitions = {
        f"location=L-0/product=P-0/{partition_suffix}": {
            "files": {"core": partition0_core}
        },
        f"location=L-1/product=P-0/{partition_suffix}": {
            "files": {"core": partition1_core}
        },
    }
    expected_indices = {
        "location": {
            "L-0": [f"location=L-0/product=P-0/{partition_suffix}"],
            "L-1": [f"location=L-1/product=P-0/{partition_suffix}"],
        },
        "product": {
            "P-0": [
                f"location=L-0/product=P-0/{partition_suffix}",
                f"location=L-1/product=P-0/{partition_suffix}",
            ]
        },
    }

    store.put(partition0_core, b"test")
    store.put(partition1_core, b"test")
    store_schema_metadata(
        make_meta(pd.DataFrame({"location": ["L-0"], "product": ["P-0"]}), origin="1"),
        dataset_uuid,
        store,
        "core",
    )

    # the following files are garbage and should not interfere with the indices and/or partitions
    for suffix in ["", ".json", ".msgpack", ".my_own_file_format"]:
        store.put(f"this_should_not_exist{suffix}", b"ignore me")
        store.put(f"{dataset_uuid}/this_should_not_exist{suffix}", b"ignore me")
        store.put(
            "{}/{}/this_should_not_exist{}".format(dataset_uuid, "core", suffix),
            b"ignore me",
        )
        store.put(
            "{}/{}/location=L-0/this_should_not_exist{}".format(
                dataset_uuid, "core", suffix
            ),
            b"ignore me",
        )

    dmd = DatasetMetadata.load_from_dict(metadata, store)
    dmd = dmd.load_partition_indices()
    dmd_dict = dmd.to_dict()
    assert dmd_dict["partitions"] == expected_partitions
    # Sorting may differ in the index list. This is ok for runtime
    # but does produce flaky tests thus sort them.
    sorted_result = {
        column: {label: sorted(x) for label, x in index.items()}
        for column, index in dmd_dict["indices"].items()
    }
    assert sorted_result == expected_indices


def test_dynamic_partitions_quote(store, metadata_version):
    """Do not specify partitions in metadata, but read them dynamically from
    store."""
    dataset_uuid = "uuid-namespace-attribute12_underscored"
    partition0_core = create_partition_key(
        dataset_uuid, "core", [("location", "München")], "data.parquet"
    )
    partition1_core = create_partition_key(
        dataset_uuid, "core", [("location", "å\\ øß")], "data.parquet"
    )
    metadata = {
        "dataset_metadata_version": metadata_version,
        "dataset_uuid": dataset_uuid,
    }
    expected_partitions = {
        "location=M%C3%BCnchen/data": {"files": {"core": partition0_core}},
        "location=%C3%A5%5C%20%C3%B8%C3%9F/data": {"files": {"core": partition1_core}},
    }
    expected_indices = {
        "location": {
            "München": ["location=M%C3%BCnchen/data"],
            "å\\ øß": ["location=%C3%A5%5C%20%C3%B8%C3%9F/data"],
        }
    }

    store.put(partition0_core, b"test")
    store.put(partition1_core, b"test")
    store_schema_metadata(
        make_meta(pd.DataFrame({"location": ["L-0"]}), origin="1"),
        dataset_uuid,
        store,
        "core",
    )

    dmd = DatasetMetadata.load_from_dict(metadata, store)
    dmd = dmd.load_partition_indices()

    dmd_dict = dmd.to_dict()
    assert dmd_dict["partitions"] == expected_partitions
    assert dmd_dict["indices"] == expected_indices


def test_dask_partitions(metadata_version, tmp_path):
    """Create partitions for one table with dask and check that it can be read
    with plateau."""
    import dask.dataframe

    dataset_uuid = "uuid+namespace-attribute12_underscored"
    (tmp_path / dataset_uuid).mkdir()
    table_path = tmp_path / dataset_uuid / "core"
    table_path.mkdir()
    store = minimalkv.get_store_from_url(f"hfs://{tmp_path}")

    locations = [f"L-{i}" for i in range(2)]
    df = pd.DataFrame()
    for location in locations:
        core = pd.DataFrame(
            data={
                "date": np.array(
                    ["2017-11-23", "2017-11-23", "2017-11-24", "2017-11-24"]
                ),
                "product": np.array(["P-0", "P-1", "P-0", "P-1"]),
                "location": location,
                "value": np.array(random.sample(range(1, 100), 4)),
            }
        )
        df = pd.concat([df, core])

    ddf = dask.dataframe.from_pandas(df, npartitions=1)
    dask_kwargs = {}
    if version.parse(dask.__version__) >= version.parse("2022.4.2"):
        dask_kwargs["write_metadata_file"] = True
    dask.dataframe.to_parquet(
        ddf, str(table_path), partition_on=["location"], **dask_kwargs
    )

    partition0 = f"{dataset_uuid}/core/location=L-0/part.0.parquet"
    partition1 = f"{dataset_uuid}/core/location=L-1/part.0.parquet"
    metadata = {
        "dataset_metadata_version": metadata_version,
        "dataset_uuid": dataset_uuid,
    }
    expected_partitions = {
        "partitions": {
            "location=L-0": {"files": {"core": partition0}},
            "location=L-1": {"files": {"core": partition1}},
        }
    }
    expected_tables = {"tables": {"core": ["date", "product", "value"]}}

    store.put(
        f"{dataset_uuid}.by-dataset-metadata.json",
        simplejson.dumps(metadata).encode(),
    )

    metadata.update(expected_partitions)
    metadata.update(expected_tables)
    dmd = DatasetMetadata.load_from_store(dataset_uuid, store)
    actual_partitions = dmd.to_dict()["partitions"]
    # we partition on location ID which has two values
    assert len(actual_partitions) == 2
    assert dmd.partition_keys == ["location"]


def test_overlap_keyspace(store, metadata_version):
    dataset_uuid1 = "uuid+namespace-attribute12_underscored"
    dataset_uuid2 = "uuid+namespace-attribute12_underscored_ext"
    table = "core"

    for dataset_uuid in (dataset_uuid1, dataset_uuid2):
        partition0 = "location=L-0"
        partition0_key = f"{dataset_uuid}/{table}/{partition0}/data.parquet"
        metadata = {
            "dataset_metadata_version": metadata_version,
            "dataset_uuid": dataset_uuid,
        }

        # put two partitions for two tables each to store
        store.put(
            f"{dataset_uuid}{naming.METADATA_BASE_SUFFIX}.json",
            simplejson.dumps(metadata).encode("utf-8"),
        )
        store.put(partition0_key, b"test")
        store_schema_metadata(
            make_meta(pd.DataFrame({"location": ["L-0"]}), origin="1"),
            dataset_uuid,
            store,
            "core",
        )

    for dataset_uuid in (dataset_uuid1, dataset_uuid2):
        partition0_label = "location=L-0/data"
        partition0_key = f"{dataset_uuid}/{table}/{partition0_label}.parquet"
        expected_partitions = {"location=L-0/data": {"files": {"core": partition0_key}}}
        expected_indices = {"location": {"L-0": ["location=L-0/data"]}}
        assert DatasetMetadata.storage_keys(dataset_uuid, store) == [
            f"{dataset_uuid}{naming.METADATA_BASE_SUFFIX}.json",
            _get_common_metadata_key(dataset_uuid, "core"),
            partition0_key,
        ]
        dmd = DatasetMetadata.load_from_store(dataset_uuid, store)
        dmd = dmd.load_partition_indices()
        dmd_dict = dmd.to_dict()
        assert dmd_dict["partitions"] == expected_partitions
        assert dmd_dict["indices"] == expected_indices



"""UUID generation mechanism used in plateau.

Using these routines ensures compatibility w/ plateau as well as the
application of best practices.
"""

import uuid


def gen_uuid() -> str:
    """Generate new UUID.

    Returns
    -------
    uuid: str
        UUID
    """
    return _uuid_hook_str()


def _uuid_hook_object():
    """Internal UUID function that could easily be overwritten for tests.

    Returns
    -------
    uuid: uuid.UUID
        UUID
    """
    return uuid.uuid4()


def _uuid_hook_str() -> str:
    """Internal UUID function that could easily be overwritten for tests.

    Returns
    -------
    uuid: str
        UUID
    """
    return _uuid_hook_object().hex


def gen_uuid_object():
    """Generate new UUID.

    Returns
    -------
    uuid: uuid.UUID
        UUID
    """
    return _uuid_hook_object()



import pickle
from itertools import permutations

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pyarrow as pa
import pyarrow.parquet as pq
import pytest
import simplejson
from dask.dataframe.utils import make_meta as dask_make_meta
from packaging.version import parse as parse_version

from plateau.core._compat import pandas_infer_string
from plateau.core.common_metadata import (
    SchemaWrapper,
    _diff_schemas,
    _get_common_metadata_key,
    empty_dataframe_from_schema,
    make_meta,
    read_schema_metadata,
    store_schema_metadata,
    validate_compatible,
    validate_shared_columns,
)
from plateau.serialization import ParquetSerializer
from plateau.serialization._parquet import PARQUET_VERSION

try:
    ARROW_VERSION = parse_version(pa.__version__)
    ARROW_DEV = ARROW_VERSION.is_devrelease
except Exception:
    ARROW_DEV = True


def test_store_schema_metadata(store, df_all_types):
    store_schema_metadata(
        schema=make_meta(df_all_types, origin="df_all_types"),
        dataset_uuid="some_uuid",
        store=store,
        table="some_table",
    )

    key = "some_uuid/some_table/_common_metadata"
    assert key in store.keys()
    pq_file = pq.ParquetFile(store.open(key))
    actual_schema = pq_file.schema.to_arrow_schema()
    # The list types hold a field which is the data type. They appear to be constructed differently
    # pa.list_(pa.field("item", pa.float64()))
    if ARROW_VERSION < parse_version("20.0.0"):
        nested_element_label = "item"
    else:
        nested_element_label = "element"
    fields = [
        pa.field(
            "array_float32", pa.list_(pa.field(nested_element_label, pa.float64()))
        ),
        pa.field(
            "array_float64", pa.list_(pa.field(nested_element_label, pa.float64()))
        ),
        pa.field("array_int16", pa.list_(pa.field(nested_element_label, pa.int64()))),
        pa.field("array_int32", pa.list_(pa.field(nested_element_label, pa.int64()))),
        pa.field("array_int64", pa.list_(pa.field(nested_element_label, pa.int64()))),
        pa.field("array_int8", pa.list_(pa.field(nested_element_label, pa.int64()))),
        pa.field("array_uint16", pa.list_(pa.field(nested_element_label, pa.uint64()))),
        pa.field("array_uint32", pa.list_(pa.field(nested_element_label, pa.uint64()))),
        pa.field("array_uint64", pa.list_(pa.field(nested_element_label, pa.uint64()))),
        pa.field("array_uint8", pa.list_(pa.field(nested_element_label, pa.uint64()))),
        pa.field(
            "array_unicode", pa.list_(pa.field(nested_element_label, pa.string()))
        ),
        pa.field("bool", pa.bool_()),
        pa.field("byte", pa.binary()),
        pa.field("date", pa.date32()),
        pa.field("datetime64", pa.timestamp("us")),
        pa.field("float32", pa.float64()),
        pa.field("float64", pa.float64()),
        pa.field("int16", pa.int64()),
        pa.field("int32", pa.int64()),
        pa.field("int64", pa.int64()),
        pa.field("int8", pa.int64()),
        pa.field("null", pa.null()),
        pa.field("uint16", pa.uint64()),
        pa.field("uint32", pa.uint64()),
        pa.field("uint64", pa.uint64()),
        pa.field("uint8", pa.uint64()),
        pa.field("unicode", pa.large_string()),
    ]
    expected_schema = pa.schema(fields)

    assert actual_schema.remove_metadata() == expected_schema


def test_schema_roundtrip(df_all_types, store):
    expected_meta = make_meta(df_all_types, origin="df_all_types")
    store_schema_metadata(
        expected_meta, dataset_uuid="dataset_uuid", store=store, table="table"
    )
    result = read_schema_metadata(
        dataset_uuid="dataset_uuid", store=store, table="table"
    )
    assert result == expected_meta


def test_pickle(df_all_types):
    obj1 = make_meta(df_all_types, origin="df_all_types")
    s = pickle.dumps(obj1)
    obj2 = pickle.loads(s)
    assert obj1 == obj2


def test_wrapper(df_all_types):
    obj = make_meta(df_all_types, origin="df_all_types")
    assert isinstance(obj, SchemaWrapper)
    assert isinstance(obj.metadata, dict)
    assert isinstance(obj.internal(), pa.Schema)

    obj2 = make_meta(df_all_types, origin="df_all_types")
    assert obj == obj2
    assert obj == obj2.internal()
    assert obj.equals(obj2)
    assert obj.equals(obj2.internal())
    assert repr(obj) == repr(obj.internal())

    assert isinstance(obj[0], pa.Field)


def test_unicode_col():
    df = pd.DataFrame({"fö": [1]})
    make_meta(df, origin="df")


def test_strip_categories():
    input_df = pd.DataFrame(
        {"categories": pd.Series(["a", "b", "c", "a"], dtype="category")}
    )

    assert len(input_df["categories"].cat.categories) > 1
    meta = make_meta(input_df, origin="input_df")
    # We strip categories to have the underlying type as the information.
    # Categories also include their categorical values as type information,
    # we're not interested in keeping them the same in all partitions.
    assert not pa.types.is_dictionary(meta[0].type)


def test_reorder(df_all_types):
    df2 = df_all_types.copy()
    df2 = df2.reindex(reversed(df_all_types.columns), axis=1)
    expected = make_meta(df_all_types, origin="df_all_types")
    actual = make_meta(df2, origin="df2")
    assert expected == actual


def test_reorder_partition_keys(df_all_types):
    partition_keys = ["int8", "uint8", "array_unicode"]
    df2 = df_all_types.copy()
    df2 = df2.reindex(reversed(df_all_types.columns), axis=1)

    expected = make_meta(
        df_all_types, origin="df_all_types", partition_keys=partition_keys
    )
    assert expected.names[: len(partition_keys)] == partition_keys

    actual = make_meta(df2, origin="df2", partition_keys=partition_keys)
    assert expected == actual


def test_compat_old_rw_path(df_all_types, store):
    # strip down DF before some column types weren't supported before anyway
    df = df_all_types[
        [
            c
            for c in df_all_types.columns
            if (
                not c.startswith("array_")  # array types (always null)
                and c != "unicode"  # unicode type (always null)
                and "8" not in c  # 8 bit types are casted to 64 bit
                and "16" not in c  # 16 bit types are casted to 64 bit
                and "32" not in c  # 32 bit types are casted to 64 bit
            )
        ]
    ]
    expected_meta = make_meta(df, origin="df")

    # old schema write path
    old_meta = dask_make_meta(df)
    pa_table = pa.Table.from_pandas(old_meta)
    buf = pa.BufferOutputStream()
    pq.write_table(pa_table, buf, version=PARQUET_VERSION)
    key_old = _get_common_metadata_key("dataset_uuid_old", "table")
    store.put(key_old, buf.getvalue().to_pybytes())

    actual_meta = read_schema_metadata(
        dataset_uuid="dataset_uuid_old", store=store, table="table"
    )
    validate_compatible([actual_meta, expected_meta])

    store_schema_metadata(
        schema=make_meta(df, origin="df"),
        dataset_uuid="dataset_uuid_new",
        store=store,
        table="table",
    )
    key_new = _get_common_metadata_key("dataset_uuid_new", "table")
    actual_df = ParquetSerializer.restore_dataframe(key=key_new, store=store)
    actual_df["date"] = actual_df["date"].dt.date
    pdt.assert_frame_equal(actual_df, old_meta)


def test_validate_compatible_same(df_all_types):
    schema1 = make_meta(df_all_types, origin="1")
    schema2 = make_meta(df_all_types, origin="2")
    schema3 = make_meta(df_all_types, origin="3")
    validate_compatible([])
    validate_compatible([schema1])
    validate_compatible([schema1, schema2])
    validate_compatible([schema1, schema2, schema3])


@pytest.mark.parametrize("remove_metadata", [True, False])
@pytest.mark.parametrize("ignore_pandas", [True, False])
def test_validate_compatible_other_pandas(df_all_types, remove_metadata, ignore_pandas):
    def _with_pandas(version):
        schema = make_meta(df_all_types, origin=version)
        metadata = schema.metadata
        pandas_metadata = simplejson.loads(metadata[b"pandas"].decode("utf8"))
        pandas_metadata["pandas_version"] = version
        metadata[b"pandas"] = simplejson.dumps(pandas_metadata).encode("utf8")
        schema = SchemaWrapper(pa.schema(schema, metadata), version)
        if remove_metadata:
            return schema.remove_metadata()
        else:
            return schema

    schema1 = make_meta(df_all_types, origin="all")
    schema2 = _with_pandas("0.19.0")
    schema3 = _with_pandas("0.99.0")
    if remove_metadata and not ignore_pandas:
        # This should fail as long as we have the metadata attached
        with pytest.raises(ValueError):
            validate_compatible(
                [schema1, schema2, schema3], ignore_pandas=ignore_pandas
            )
        schema1 = schema1.remove_metadata()
    validate_compatible([schema1, schema2, schema3], ignore_pandas=ignore_pandas)


def test_validate_compatible_different(df_all_types):
    df2 = df_all_types.loc[:, df_all_types.columns[:2]].copy()
    schema1 = make_meta(df_all_types, origin="1")
    schema2 = make_meta(df2, origin="2")
    with pytest.raises(ValueError) as exc:
        validate_compatible([schema1, schema2])
    assert str(exc.value).startswith("Schema violation")


def test_validate_shared_columns_same(df_all_types):
    schema1 = make_meta(df_all_types, origin="1")
    schema2 = make_meta(df_all_types, origin="2")
    schema3 = make_meta(df_all_types, origin="3").remove_metadata()
    validate_shared_columns([])
    validate_shared_columns([schema1])
    validate_shared_columns([schema1, schema2])
    with pytest.raises(ValueError):
        validate_shared_columns([schema1, schema2, schema3])
    validate_shared_columns([schema1, schema2, schema3], ignore_pandas=True)
    validate_shared_columns(
        [schema1.remove_metadata(), schema2.remove_metadata(), schema3]
    )


def test_validate_shared_columns_null_value(df_all_types):
    schema1 = make_meta(df_all_types, origin="1")
    schema2 = make_meta(df_all_types.drop(0), origin="2")
    schema3 = make_meta(df_all_types, origin="3").remove_metadata()
    validate_shared_columns([])
    validate_shared_columns([schema1])
    validate_shared_columns([schema1, schema2])
    with pytest.raises(ValueError):
        validate_shared_columns([schema1, schema2, schema3])
    validate_shared_columns([schema1, schema2, schema3], ignore_pandas=True)
    validate_shared_columns(
        [schema1.remove_metadata(), schema2.remove_metadata(), schema3]
    )


def test_validate_shared_columns_no_share(df_all_types):
    schema1 = make_meta(df_all_types.loc[:, df_all_types.columns[0:2]], origin="1")
    schema2 = make_meta(df_all_types.loc[:, df_all_types.columns[2:4]], origin="2")
    schema3 = make_meta(df_all_types.loc[:, df_all_types.columns[4:6]], origin="3")
    validate_shared_columns([])
    validate_shared_columns([schema1])
    validate_shared_columns([schema1, schema2])
    validate_shared_columns([schema1, schema2, schema3])


@pytest.mark.parametrize("remove_metadata", [True, False])
def test_validate_shared_columns_fail(df_all_types, remove_metadata):
    df2 = df_all_types.copy()
    df2["uint16"] = df2["uint16"].astype(float)
    schema1 = make_meta(df_all_types, origin="1")
    schema2 = make_meta(df2, origin="2")
    if remove_metadata:
        schema1 = schema1.remove_metadata()
        schema2 = schema2.remove_metadata()
    with pytest.raises(ValueError) as exc:
        validate_shared_columns([schema1, schema2])
    assert str(exc.value).startswith('Found incompatible entries for column "uint16"')


def test_validate_empty_dataframe(
    df_all_types, df_all_types_schema, df_all_types_empty_schema
):
    # Do not raise in case one of the schemas is of an empty dataframe
    # Test all permutations to avoid that the implementation is sensitive on whether
    # the first schema is empty/non-empty
    for schemas in permutations([df_all_types_schema, df_all_types_empty_schema]):
        validate_compatible(schemas)
    validate_compatible([df_all_types_empty_schema, df_all_types_empty_schema])


@pytest.mark.parametrize(
    "corrupt_column,corrupt_value,corrupt_dtype",
    [
        # reference column is a native type
        ("int8", -1.1, np.float64),
        ("int8", "a", object),
        # reference column is an object
        ("unicode", -1.1, np.float64),
        ("unicode", 1, np.int64),
        pytest.param(
            "unicode",
            None,
            None,
            marks=pytest.mark.xfail(
                strict=True,
                reason="This results in a `null` column which cannot be compared and must not fail",
            ),
        ),
        # reference column is a native typed array
        ("array_int64", np.array([1.0], dtype=np.float64), object),
        ("array_int64", np.array(["a"], dtype=object), object),
        # reference column is an object types arrayarray
        ("array_unicode", np.array([1], dtype=np.int8), object),
        ("array_unicode", np.array([1.0], dtype=np.float64), object),
    ],
)
def test_validate_empty_dataframe_corrupt_raises(
    df_all_types,
    df_all_types_schema,
    df_all_types_empty_schema,
    corrupt_column,
    corrupt_value,
    corrupt_dtype,
):
    # In case there is something wrong with the schema, raise!

    # First, an integer column carries a float or an object.
    df_corrupt = df_all_types.copy()
    # for value, dtype in [(-1.1, np.float64), ('a', object)]:
    df_corrupt[corrupt_column] = pd.Series([corrupt_value], dtype=corrupt_dtype)
    df_corrupt_meta = make_meta(df_corrupt, origin="1")
    # Raise when comparing the proper to the corrupt schema
    for schemas in permutations([df_all_types_schema, df_corrupt_meta]):
        with pytest.raises(ValueError):
            validate_compatible(schemas)
    # Also raise if there is a schema originating from an empty DF to make
    # sure the emptiness doesn't cancel the validation
    for schemas in permutations(
        [df_all_types_schema, df_corrupt_meta, df_all_types_empty_schema]
    ):
        with pytest.raises(ValueError):
            validate_compatible(schemas)


def test_validate_different_cats_same_type():
    input_df = pd.DataFrame(
        {"categories": pd.Series(["a", "b", "c", "a"], dtype="category")}
    )
    input_df_2 = pd.DataFrame(
        {"categories": pd.Series(["f", "e", "e", "f"], dtype="category")}
    )
    input_df_3 = pd.DataFrame({"categories": pd.Series(["f", "e", "e", "f"])})

    meta = make_meta(input_df, origin="1")
    meta_2 = make_meta(input_df_2, origin="2")
    meta_3 = make_meta(input_df_3, origin="3")
    validate_compatible([meta, meta_2, meta_3])


def test_validate_different_cats_different_type():
    input_df = pd.DataFrame(
        {"categories": pd.Series(["a", "b", "c", "a"], dtype="category")}
    )
    input_df_2 = pd.DataFrame(
        {"categories": pd.Series([b"f", b"e", b"e", b"f"], dtype="category")}
    )

    meta = make_meta(input_df, origin="1")
    meta_2 = make_meta(input_df_2, origin="2")
    with pytest.raises(ValueError):
        validate_compatible([meta, meta_2])


@pytest.mark.parametrize(
    "index", [pd.Index([0], dtype="int64"), pd.RangeIndex(start=0, stop=1)]
)
def test_schema_dataframe_roundtrip(index, df_all_types):
    df = pd.DataFrame(df_all_types, index=index)

    schema = make_meta(df, origin="1")
    actual_df = empty_dataframe_from_schema(schema, date_as_object=True)
    validate_compatible([schema, make_meta(actual_df, origin="2")])


def test_empty_dataframe_from_schema(df_all_types):
    schema = make_meta(df_all_types, origin="1")
    actual_df = empty_dataframe_from_schema(schema)

    expected_df = df_all_types.loc[[]]
    expected_df["date"] = pd.Series([], dtype="datetime64[ns]")
    for c in expected_df.columns:
        if c.startswith("float"):
            expected_df[c] = pd.Series([], dtype=float)
        if c.startswith("int"):
            expected_df[c] = pd.Series([], dtype=int)
        if c.startswith("uint"):
            expected_df[c] = pd.Series([], dtype=np.uint64)

    pdt.assert_frame_equal(actual_df, expected_df)


def test_empty_dataframe_from_schema_columns(df_all_types):
    schema = make_meta(df_all_types, origin="1")
    actual_df = empty_dataframe_from_schema(schema, ["uint64", "int64"])

    expected_df = df_all_types.loc[[], ["uint64", "int64"]]
    pdt.assert_frame_equal(actual_df, expected_df)


@pytest.mark.xfail(ARROW_DEV, reason="Format not stbale")
def test_diff_schemas(df_all_types):
    # Prepare a schema with one missing, one additional and one changed column
    df2 = df_all_types.drop(columns=df_all_types.columns[0])
    df2["new_col"] = pd.Series(df_all_types["bool"])
    df2["int16"] = df2["int16"].astype(float)
    df2 = df2.reset_index(drop=True)

    schema2 = make_meta(df2, origin="2")
    schema1 = make_meta(df_all_types, origin="1")
    diff = _diff_schemas(schema1, schema2)
    expected_arrow_diff = """Arrow schema:
@@ -1,5 +1,3 @@

-array_float32: list<item: double>
-  child 0, item: double
 array_float64: list<item: double>
   child 0, item: double
 array_int16: list<item: int64>
@@ -26,10 +24,11 @@

 datetime64: timestamp[ns]
 float32: double
 float64: double
-int16: int64
+int16: double
 int32: int64
 int64: int64
 int8: int64
+new_col: bool
 null: null
 uint16: uint64
 uint32: uint64

"""
    expected_pandas_diff = (
        f"""Pandas_metadata:
@@ -3,12 +3,7 @@

                      'name': None,
                      'numpy_type': '{"str" if pandas_infer_string() else "object"}',"""
        + """
                      'pandas_type': 'unicode'}],
- 'columns': [{'field_name': 'array_float32',
-              'metadata': None,
-              'name': 'array_float32',
-              'numpy_type': 'object',
-              'pandas_type': 'list[float64]'},
-             {'field_name': 'array_float64',
+ 'columns': [{'field_name': 'array_float64',
               'metadata': None,
               'name': 'array_float64',
               'numpy_type': 'object',
@@ -91,8 +86,8 @@

              {'field_name': 'int16',
               'metadata': None,
               'name': 'int16',
-              'numpy_type': 'int64',
-              'pandas_type': 'int64'},
+              'numpy_type': 'float64',
+              'pandas_type': 'float64'},
              {'field_name': 'int32',
               'metadata': None,
               'name': 'int32',
@@ -108,6 +103,11 @@

               'name': 'int8',
               'numpy_type': 'int64',
               'pandas_type': 'int64'},
+             {'field_name': 'new_col',
+              'metadata': None,
+              'name': 'new_col',
+              'numpy_type': 'bool',
+              'pandas_type': 'bool'},
              {'field_name': 'null',
               'metadata': None,
               'name': 'null',"""
    )

    assert diff == expected_arrow_diff + expected_pandas_diff


def test_validate_schema_non_overlapping_nulls(df_all_types_schema):
    """Test that two schemas with non-overlapping null columns are valid."""
    first_ix = np.random.randint(len(df_all_types_schema))
    second_ix = first_ix
    while second_ix == first_ix:
        second_ix = np.random.randint(len(df_all_types_schema))

    first_null = pa.field(name=df_all_types_schema.names[first_ix], type=pa.null())
    first_schema = df_all_types_schema.set(first_ix, first_null)

    second_null = pa.field(name=df_all_types_schema.names[second_ix], type=pa.null())
    second_schema = df_all_types_schema.set(second_ix, second_null)

    for schemas in permutations([first_schema, second_schema]):
        reference_schema = validate_compatible(schemas)

        # The reference schema should be the original schema
        # with the columns reconstructed
        assert df_all_types_schema == reference_schema


def test_make_meta_column_normalization_pyarrow_schema():
    # GH228
    df = pd.DataFrame(
        [{"part": 1, "id": 1, "col1": "abc"}, {"part": 2, "id": 2, "col1": np.nan}],
        # plateau normalizes field order s.t. partition keys are first and the
        # rest is alphabetically. This is reverse.
        columns=["col1", "id", "part"],
    )
    schema = make_meta(
        pa.Schema.from_pandas(df), origin="gh228", partition_keys=["part"]
    )
    fields = [
        pa.field("part", pa.int64()),
        pa.field("col1", pa.large_string()),
        pa.field("id", pa.int64()),
    ]
    expected_schema = pa.schema(fields)

    assert schema.internal().equals(expected_schema, check_metadata=False)



import pytest

from plateau.io.eager import build_dataset_indices
from plateau.io.testing.index import *  # noqa: F403, F4


@pytest.fixture()
def bound_build_dataset_indices():
    return build_dataset_indices



# add this to conf.py

from sphinx.addnodes import pending_xref
from sphinx.ext.intersphinx import missing_reference

# https://stackoverflow.com/questions/62293058/how-to-add-objects-to-sphinxs-global-index-or-cross-reference-by-alias


def replace_target_intersphinx(app, env, node, contnode):
    reftarget_replace = app.config.reftarget_replace
    target = node.get("reftarget", None)
    if target:
        for old, new in reftarget_replace.items():
            node["reftarget"] = target.replace(old, new)
    return missing_reference(app, env, node, contnode)


def replace_target(app, doctree):
    reftarget_replace = app.config.reftarget_replace
    pending_xrefs = doctree.traverse(condition=pending_xref)

    for node in pending_xrefs:
        for old, new in reftarget_replace.items():
            target = node.get("reftarget")
            if target:
                new_target = target.replace(old, new)
                node["reftarget"] = new_target


def setup(app):
    app.add_config_value("reftarget_replace", default={}, rebuild="env")
    app.connect("doctree-read", replace_target)
    app.connect("missing-reference", replace_target_intersphinx)



"""This module contains functionality for persisting/serialising DataFrames."""

import datetime
import logging
import time
from collections.abc import Iterable
from typing import Any

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from minimalkv import KeyValueStore
from pyarrow.parquet import ParquetFile

from ._generic import (
    ConjunctionType,
    DataFrameSerializer,
    PredicatesType,
    check_predicates,
    filter_df,
    filter_df_from_predicates,
)
from ._io_buffer import BlockBuffer
from ._util import ensure_unicode_string_type, schema_metadata_bytes_to_object

try:
    # Only check for BotoStore instance if boto is really installed
    from minimalkv.net.botostore import BotoStore

    HAVE_BOTO = True
except ImportError:
    HAVE_BOTO = False

_logger = logging.getLogger(__name__)


EPOCH_ORDINAL = datetime.date(1970, 1, 1).toordinal()
MAX_NB_RETRIES = 6  # longest retry backoff = BACKOFF_TIME * 2**(MAX_NB_RETRIES - 2)
BACKOFF_TIME = 0.01  # 10 ms
PARQUET_VERSION = "2.4"


def _empty_table_from_schema(parquet_file):
    schema = parquet_file.schema.to_arrow_schema()

    return schema.empty_table()


def _reset_dictionary_columns(table, exclude=None):
    """We need to ensure that the dtype is exactly as requested, see GH227."""
    from plateau.core._compat import ARROW_GE_20
    from plateau.core.common_metadata import _dict_to_binary

    if exclude is None:
        exclude = []

    schema = table.schema
    pandas_metadata = schema.pandas_metadata
    for i in range(len(schema)):
        field = schema[i]
        if field.name in exclude:
            continue
        if pa.types.is_dictionary(field.type):
            new_field = pa.field(
                field.name,
                field.type.value_type,
                field.nullable,
                field.metadata,
            )
            schema = schema.remove(i).insert(i, new_field)
            if not ARROW_GE_20:
                pandas_metadata["columns"][i]["pandas_type"] = "object"
                pandas_metadata["columns"][i]["numpy_type"] = "str"
    schema = schema.remove_metadata()
    md = {b"pandas": _dict_to_binary(pandas_metadata)}
    schema = schema.with_metadata(md)
    table = table.cast(schema)
    return table


class ParquetReadError(IOError):
    """Internal plateau error while attempting to read Parquet file."""

    pass


class ParquetSerializer(DataFrameSerializer):
    """Serializer to store a :class:`pandas.DataFrame` as parquet.

    On top of the plain serialization, this class handles forward and
    backwards compatibility between pyarrow versions.

    Parameters
    ----------
    compression
        The compression algorithm to be used for the parquet file. For a
        comprehensive list of available compression algorithms, please
        see :func:`pyarrow.parquet.write_table`.
        The default is set to "SNAPPY" which usually offers a good balance
        between performance and compression rate. Depending on your data,
        picking a different algorithm may have vastly different
        characteristics and we can only recommend to test this on your own
        data. Depending on the reader parquet implementation, some
        compression algorithms may not be supported and we recommend to
        consult the documentation of the reader libraries first.
    chunk_size
        The number of rows stored in a Parquet RowGroup. To leverage
        predicate pushdown, it is necessary to set this value. We do not
        apply any default value since a good choice is very sensitive to the
        kind of data you are using and what kind of storage.
        A typical range to try out would be somewhere between 50k-200k. To fully leverage row group statistics, it is highly recommended to sort the file before serialization.

    Notes
    -----

    Regarding type stability and supported types there are a few known limitations users should be aware of.


    .. ipython:: python
        :suppress:

        from plateau.core.utils import ensure_store
        import pandas as pd
        from plateau.serialization import ParquetSerializer

        store = ensure_store("hmemory://")

    * `pandas.Categorical`

        plateau offers the keyword argument `categories` which contains a list of field names which are supposed to retrieved as a `pandas.Categorical`.

        See also :ref:`Dictionary Encoding`

        .. ipython:: python
            :okwarning:

            ser = ParquetSerializer()

            df = pd.DataFrame({"cat_field": pd.Categorical(["A"])})
            df.dtypes
            ser.restore_dataframe(store, ser.store(store, "cat", df))
            ser.restore_dataframe(store, ser.store(store, "cat", df), categories=["cat_field"])

    * Timestamps with nanosecond resolution

        Timestamps can only be stored in micro second (`us`) accuracy. Trying to do differently may raise an exception.

        See also :ref:`timestamp`

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyarrow as pa

            pa.__version__

            df = pd.DataFrame({"nanosecond": [pd.Timestamp("2021-01-01 00:00:00.0000001")]})
            # nanosecond resolution
            ser.store(store, "key", df)
    """

    type_stable = True

    def __init__(
        self, compression: str = "SNAPPY", chunk_size: int | None = None
    ) -> None:
        self.compression = compression

        if chunk_size is not None:
            if not isinstance(chunk_size, int):
                raise TypeError(
                    "Cannot initialize ParquetSerializer because chunk size is not integer type"
                )
            if chunk_size < 1:
                raise ValueError(
                    "Cannot initialize ParquetSerializer because chunk size < 1"
                )
        self.chunk_size = chunk_size

    def __eq__(self, other):
        return (
            isinstance(other, ParquetSerializer)
            and (self.compression == other.compression)
            and (self.chunk_size == other.chunk_size)
        )

    def __repr__(self):
        return f"ParquetSerializer(compression={self.compression!r}, chunk_size={self.chunk_size!r})"

    @staticmethod
    def _restore_dataframe(
        store: KeyValueStore,
        key: str,
        filter_query: str | None = None,
        columns: Iterable[str] | None = None,
        predicate_pushdown_to_io: bool = True,
        categories: Iterable[str] | None = None,
        predicates: PredicatesType | None = None,
        date_as_object: bool = False,
    ) -> pd.DataFrame:
        check_predicates(predicates)
        _coerce = {"coerce_temporal_nanoseconds": True}

        # If we want to do columnar access we can benefit from partial reads
        # otherwise full read en block is the better option.
        if (not predicate_pushdown_to_io) or (columns is None and predicates is None):
            with pa.BufferReader(store.get(key)) as reader:
                table = pq.read_pandas(reader, columns=columns)
        else:
            if HAVE_BOTO and isinstance(store, BotoStore):
                # Parquet and seeks on S3 currently leak connections thus
                # we omit column projection to the store.
                reader = pa.BufferReader(store.get(key))
            else:
                reader = store.open(key)
                # Buffer at least 4 MB in requests. This is chosen because the default block size of the Azure
                # storage client is 4MB.
                reader = BlockBuffer(reader, 4 * 1024 * 1024)
            try:
                parquet_file = ParquetFile(reader)

                # Check earlier for missing columns to produce the same error
                # as we ever did with earlier pyarrow versions
                if columns is not None:
                    missing_columns = set(columns) - set(
                        parquet_file.schema.to_arrow_schema().names
                    )
                    if missing_columns:
                        raise ValueError(
                            "Columns cannot be found in stored dataframe: {missing}".format(
                                missing=", ".join(sorted(missing_columns))
                            )
                        )

                if predicates and parquet_file.metadata.num_rows > 0:
                    # We need to calculate different predicates for predicate
                    # pushdown and the later DataFrame filtering. This is required
                    # e.g. in the case where we have an `in` predicate as this has
                    # different normalized values.
                    columns_to_io = _columns_for_pushdown(columns, predicates)
                    predicates_for_pushdown = _normalize_predicates(
                        parquet_file, predicates, True
                    )
                    predicates = _normalize_predicates(parquet_file, predicates, False)
                    tables = _read_row_groups_into_tables(
                        parquet_file, columns_to_io, predicates_for_pushdown
                    )

                    if len(tables) == 0:
                        table = _empty_table_from_schema(parquet_file)
                    else:
                        table = pa.concat_tables(tables)
                else:
                    # ARROW-5139 Column projection with empty columns returns a table w/out index
                    if columns == []:
                        # Create an arrow table with expected index length.
                        df = (
                            parquet_file.schema.to_arrow_schema()
                            .empty_table()
                            .to_pandas(date_as_object=date_as_object, **_coerce)
                        )
                        index = pd.Index(
                            pd.RangeIndex(start=0, stop=parquet_file.metadata.num_rows),
                            dtype="int64",
                        )
                        df = pd.DataFrame(df, index=index)
                        # convert back to table to keep downstream code untouched by this patch
                        table = pa.Table.from_pandas(df)
                    else:
                        table = pq.read_pandas(reader, columns=columns)
            finally:
                reader.close()

        if columns is not None:
            missing_columns = set(columns) - set(table.schema.names)
            if missing_columns:
                raise ValueError(
                    "Columns cannot be found in stored dataframe: {missing}".format(
                        missing=", ".join(sorted(missing_columns))
                    )
                )

        table = _reset_dictionary_columns(table, exclude=categories)

        # HACK: Cast bytes to object in metadata until Pandas bug is fixed: https://github.com/pandas-dev/pandas/issues/50127
        table = table.cast(schema_metadata_bytes_to_object(table.schema))

        df = table.to_pandas(date_as_object=date_as_object, **_coerce)

        # XXX: Patch until Pyarrow bug is resolved: https://github.com/apache/arrow/issues/33297
        if categories:
            for col in categories:
                if col in df:
                    df[col] = df[col].astype("category")

        df.columns = df.columns.map(ensure_unicode_string_type)
        if predicates:
            df = filter_df_from_predicates(
                df, predicates, strict_date_types=date_as_object
            )
        else:
            df = filter_df(df, filter_query)
        if columns is not None:
            return df.reindex(columns=columns, copy=False)
        else:
            return df

    @classmethod
    def restore_dataframe(
        cls,
        store: KeyValueStore,
        key: str,
        filter_query: str | None = None,
        columns: Iterable[str] | None = None,
        predicate_pushdown_to_io: bool = True,
        categories: Iterable[str] | None = None,
        predicates: PredicatesType | None = None,
        date_as_object: bool = False,
    ) -> pd.DataFrame:
        # https://github.com/JDASoftwareGroup/plateau/issues/407  We have been seeing weird `IOError`s while reading
        # Parquet files from Azure Blob Store. These errors have caused long running computations to fail.
        # The workaround is to retry the serialization here and gain more stability for long running tasks.
        # This code should not live forever, it should be removed once the underlying cause has been resolved.
        for nb_retry in range(MAX_NB_RETRIES):
            try:
                return cls._restore_dataframe(
                    store=store,
                    key=key,
                    filter_query=filter_query,
                    columns=columns,
                    predicate_pushdown_to_io=predicate_pushdown_to_io,
                    categories=categories,
                    predicates=predicates,
                    date_as_object=date_as_object,
                )
            # We only retry OSErrors (note that IOError inherits from OSError), as these kind of errors may benefit
            # from retries.
            except OSError as err:
                raised_error = err
                _logger.warning(
                    msg=(
                        f"Failed to restore dataframe, attempt {nb_retry + 1} of {MAX_NB_RETRIES} with parameters "
                        f"key: {key}, filter_query: {filter_query}, columns: {columns}, "
                        f"predicate_pushdown_to_io: {predicate_pushdown_to_io}, categories: {categories}, "
                        f"predicates: {predicates}, date_as_object: {date_as_object}."
                    ),
                    exc_info=True,
                )
                # we don't sleep when we're done with the last attempt
                if nb_retry < (MAX_NB_RETRIES - 1):
                    time.sleep(BACKOFF_TIME * 2**nb_retry)

        raise ParquetReadError(
            f"Failed to restore dataframe after {MAX_NB_RETRIES} attempts. Parameters: "
            f"key: {key}, filter_query: {filter_query}, columns: {columns}, "
            f"predicate_pushdown_to_io: {predicate_pushdown_to_io}, categories: {categories}, "
            f"date_as_object: {date_as_object}, predicates: {predicates}."
        ) from raised_error

    def store(self, store, key_prefix, df):
        key = f"{key_prefix}.parquet"
        if isinstance(df, pa.Table):
            table = df
        else:
            table = pa.Table.from_pandas(df)
        buf = pa.BufferOutputStream()

        pq.write_table(
            table,
            buf,
            version=PARQUET_VERSION,
            chunk_size=self.chunk_size,
            compression=self.compression,
            coerce_timestamps="us",
        )
        store.put(key, buf.getvalue().to_pybytes())
        return key


def _columns_for_pushdown(columns, predicates):
    if columns is None:
        return
    new_cols = columns[:]
    for conjunction in predicates:
        for literal in conjunction:
            if literal[0] not in columns:
                new_cols.append(literal[0])
    return new_cols


def _read_row_groups_into_tables(parquet_file, columns, predicates_in):
    """For each RowGroup check if the predicate in DNF applies and then read
    the respective RowGroup."""
    arrow_schema = parquet_file.schema.to_arrow_schema()
    parquet_reader = parquet_file.reader

    def all_predicates_accept(row):
        # Check if the predicates evaluate on this RowGroup.
        # As the predicate is in DNF, we only need a single of the
        # inner lists to match. Once we have found a positive match,
        # there is no need to check whether the remaining ones apply.
        row_meta = parquet_file.metadata.row_group(row)
        for predicate_list in predicates_in:
            if all(
                _predicate_accepts(predicate, row_meta, arrow_schema, parquet_reader)
                for predicate in predicate_list
            ):
                return True
        return False

    # Iterate over the RowGroups and evaluate the list of predicates on each
    # one of them. Only access those that could contain a row where we could
    # get an exact match of the predicate.
    result = []
    for row in range(parquet_file.num_row_groups):
        if all_predicates_accept(row):
            row_group = parquet_file.read_row_group(row, columns=columns)
            result.append(row_group)
    return result


def _normalize_predicates(
    parquet_file, predicates: list[ConjunctionType], for_pushdown
):
    schema = parquet_file.schema.to_arrow_schema()

    normalized_predicates = []
    for conjunction in predicates:
        evaluates_to_false = False
        new_conjunction: list[Any] = []

        for literal in conjunction:
            col, op, val = literal
            col_idx = parquet_file.reader.column_name_idx(col)
            pa_type = schema[col_idx].type
            column_name = schema[col_idx].name

            if pa.types.is_null(pa_type):
                # early exit, the entire conjunction evaluates to False
                evaluates_to_false = True
                break

            if op == "in":
                normalized_value = [
                    _normalize_value(lit, pa_type, column_name=column_name)
                    for lit in literal[2]
                ]
            else:
                normalized_value = _normalize_value(
                    literal[2], pa_type, column_name=column_name
                )
            new_literal = (literal[0], literal[1], normalized_value)
            new_conjunction.append(new_literal)

        if not evaluates_to_false:
            normalized_predicates.append(new_conjunction)
    return normalized_predicates


def _normalize_value(value, pa_type, column_name=None):
    if pa.types.is_dictionary(pa_type):
        pa_type = pa_type.value_type

    if pa.types.is_string(pa_type) or pa.types.is_large_string(pa_type):
        if isinstance(value, bytes):
            return value.decode("utf-8")
        elif isinstance(value, str):
            return value
        elif value is None:
            return value
    elif pa.types.is_binary(pa_type):
        if isinstance(value, bytes):
            return value
        elif isinstance(value, str):
            return str(value).encode("utf-8")
    elif (
        pa.types.is_integer(pa_type)
        and pd.api.types.is_integer(value)
        or pa.types.is_floating(pa_type)
        and pd.api.types.is_float(value)
        or pa.types.is_boolean(pa_type)
        and pd.api.types.is_bool(value)
        or pa.types.is_timestamp(pa_type)
        and not isinstance(value, bytes | str)
        and (
            pd.api.types.is_datetime64_dtype(value)
            or isinstance(value, datetime.datetime)
        )
    ):
        return value
    elif pa.types.is_date(pa_type):
        if isinstance(value, str):
            return datetime.datetime.strptime(value, "%Y-%m-%d").date()
        elif isinstance(value, bytes):
            value = value.decode("utf-8")
            return datetime.datetime.strptime(value, "%Y-%m-%d").date()
        elif isinstance(value, datetime.date):
            if isinstance(value, datetime.datetime):
                raise TypeError(
                    f"Unexpected type for predicate: Column {column_name!r} is an "
                    f"Arrow date ({pa_type}), but predicate value has type {type(value)}. "
                    f"Use a Python 'datetime.date' object instead."
                )
            else:
                return value
    predicate_value_dtype = pd.Series(value).dtype
    raise TypeError(
        f"Unexpected type for predicate: Column {column_name!r} has pandas type "
        f"{pa_type.to_pandas_dtype()} (Arrow type {pa_type}), but predicate value "
        f"{value!r} has pandas type '{predicate_value_dtype}' (Python type '{type(value)}')"
    )


def _predicate_accepts(predicate, row_meta, arrow_schema, parquet_reader):
    """Checks if a predicate evaluates on a column.

    This method first casts the value of the predicate to the type used
    for this column in the statistics and then applies the relevant
    operator. The operation applied here is done in a fashion to check
    if the predicate would evaluate to True for any possible row in the
    RowGroup. Thus e.g. for the `==` predicate, we check if the
    predicate value is in the (min, max) range of the RowGroup.
    """
    col, op, val = predicate
    col_idx = parquet_reader.column_name_idx(col)
    pa_type = arrow_schema[col_idx].type
    parquet_statistics = row_meta.column(col_idx).statistics

    # In case min/max is not set, we have to assume that the predicate matches.
    if not parquet_statistics.has_min_max:
        return True

    min_value = parquet_statistics.min
    max_value = parquet_statistics.max
    # Transform the predicate value to the respective type used in the statistics.

    # integer overflow protection since statistics are stored as signed integer, see ARROW-5166
    if pa.types.is_integer(pa_type) and (max_value < min_value):
        return True

    if pa.types.is_timestamp(pa_type):
        # timestamps in the parquet statistic might be of type datetime.datetime, which is not compatible w/ numpy
        min_value = np.datetime64(min_value)
        max_value = np.datetime64(max_value)

    # The statistics for floats only contain the 6 most significant digits.
    # So a suitable epsilon has to be considered below min and above max.
    if isinstance(val, float):
        min_value -= _epsilon(min_value)
        max_value += _epsilon(max_value)

    # op can only be "==" or "!=" for scalar null values.
    if op == "==":
        if pd.isnull(val):
            return parquet_statistics.null_count > 0
        else:
            return (min_value <= val) and (max_value >= val)
    elif op == "!=":
        if pd.isnull(val):
            return parquet_statistics.null_count < row_meta.num_rows
        else:
            return not ((min_value >= val) and (max_value <= val))
    elif op == "<=":
        return min_value <= val
    elif op == ">=":
        return max_value >= val
    elif op == "<":
        return min_value < val
    elif op == ">":
        return max_value > val
    elif op == "in":
        # This implementation is chosen for performance reasons. See
        # https://github.com/JDASoftwareGroup/kartothek/pull/130 for more information/benchmarks.
        # We accept the predicate if there is any value in the provided array which is equal to or between
        # the parquet min and max statistics. Otherwise, it is rejected.
        for x in val:
            if pd.isnull(x):
                if parquet_statistics.null_count > 0:
                    return True
            elif min_value <= x <= max_value:
                return True
        return False
    else:
        raise NotImplementedError("op not supported")


def _highest_significant_position(num):
    """
    >>> _highest_significant_position(1.0)
    1
    >>> _highest_significant_position(9.0)
    1
    >>> _highest_significant_position(39.0)
    2
    >>> _highest_significant_position(0.1)
    -1
    >>> _highest_significant_position(0.9)
    -1
    >>> _highest_significant_position(0.000123)
    -4
    >>> _highest_significant_position(1234567.0)
    7
    >>> _highest_significant_position(-0.1)
    -1
    >>> _highest_significant_position(-100.0)
    3
    """
    abs_num = np.absolute(num)
    log_of_abs = np.log10(abs_num)
    position = int(np.floor(log_of_abs))

    # is position left of decimal point?
    if abs_num >= 1.0:
        position += 1

    return position


def _epsilon(num):
    """
    >>> _epsilon(123456)
    1
    >>> _epsilon(0.123456)
    1e-06
    >>> _epsilon(0.123)
    1e-06
    >>> _epsilon(0)
    0
    >>> _epsilon(-0.123456)
    1e-06
    >>> _epsilon(-123456)
    1
    >>> _epsilon(np.inf)
    0
    >>> _epsilon(-np.inf)
    0
    """
    SIGNIFICANT_DIGITS = 6

    if num == 0 or np.isinf(num):
        return 0

    epsilon_position = _highest_significant_position(num) - SIGNIFICANT_DIGITS

    # is position right of decimal point?
    if epsilon_position < 0:
        epsilon_position += 1

    return 10**epsilon_position



from collections import OrderedDict
from datetime import date

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pyarrow as pa
import pytest

from plateau.core.common_metadata import make_meta, store_schema_metadata
from plateau.core.index import ExplicitSecondaryIndex
from plateau.core.naming import DEFAULT_METADATA_VERSION
from plateau.core.testing import get_dataframe_not_nested
from plateau.io_components.metapartition import (
    SINGLE_TABLE,
    MetaPartition,
    _unique_label,
    parse_input_to_metapartition,
    partition_labels_from_mps,
)
from plateau.serialization import DataFrameSerializer, ParquetSerializer


def test_store_single_dataframe_as_partition(store, metadata_version):
    df = pd.DataFrame(
        {"P": np.arange(0, 10), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )
    mp = MetaPartition(label="test_label", data=df, metadata_version=metadata_version)

    meta_partition = mp.store_dataframes(
        store=store, df_serializer=ParquetSerializer(), dataset_uuid="dataset_uuid"
    )

    assert meta_partition.data is None

    expected_key = "dataset_uuid/table/test_label.parquet"

    assert meta_partition.file == expected_key
    assert meta_partition.label == "test_label"

    files_in_store = list(store.keys())

    expected_num_files = 1
    assert len(files_in_store) == expected_num_files
    stored_df = DataFrameSerializer.restore_dataframe(store=store, key=expected_key)
    pdt.assert_frame_equal(df, stored_df)
    files_in_store.remove(expected_key)
    assert len(files_in_store) == expected_num_files - 1


def test_load_dataframe_logical_conjunction(store, metadata_version):
    df = pd.DataFrame(
        {"P": np.arange(0, 10), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )
    mp = MetaPartition(
        label="cluster_1",
        data=df,
        metadata_version=metadata_version,
        logical_conjunction=[("P", ">", 4)],
    )
    meta_partition = mp.store_dataframes(
        store=store, df_serializer=None, dataset_uuid="dataset_uuid"
    )
    predicates = None
    loaded_mp = meta_partition.load_dataframes(store=store, predicates=predicates)
    data = pd.DataFrame(
        {"P": [5, 6, 7, 8, 9], "L": [5, 6, 7, 8, 9], "TARGET": [15, 16, 17, 18, 19]}
    ).set_index(np.arange(5, 10))
    pdt.assert_frame_equal(loaded_mp.data, data)

    predicates = [[("L", ">", 6), ("TARGET", "<", 18)]]
    loaded_mp = meta_partition.load_dataframes(store=store, predicates=predicates)
    data = pd.DataFrame({"P": [7], "L": [7], "TARGET": [17]}).set_index(np.array([7]))
    pdt.assert_frame_equal(loaded_mp.data, data)

    predicates = [[("L", ">", 2), ("TARGET", "<", 17)], [("TARGET", "==", 19)]]
    loaded_mp = meta_partition.load_dataframes(store=store, predicates=predicates)
    data = pd.DataFrame(
        {"P": [5, 6, 9], "L": [5, 6, 9], "TARGET": [15, 16, 19]}
    ).set_index(np.array([5, 6, 9]))
    pdt.assert_frame_equal(loaded_mp.data, data)


@pytest.mark.parametrize("predicate_pushdown_to_io", [True, False])
def test_load_dataframes(
    meta_partitions_files_only, store_session, predicate_pushdown_to_io
):
    expected_df = pd.DataFrame(
        OrderedDict(
            [("P", [1]), ("L", [1]), ("TARGET", [1]), ("DATE", [date(2010, 1, 1)])]
        )
    )
    mp = meta_partitions_files_only[0]
    assert mp.file
    assert mp.data is not None
    mp = meta_partitions_files_only[0].load_dataframes(
        store=store_session, predicate_pushdown_to_io=predicate_pushdown_to_io
    )
    assert mp.data is not None
    data = mp.data

    pdt.assert_frame_equal(data, expected_df, check_dtype=False)

    empty_mp = MetaPartition("empty_mp", metadata_version=mp.metadata_version)
    empty_mp.load_dataframes(
        store_session, predicate_pushdown_to_io=predicate_pushdown_to_io
    )
    assert empty_mp.data is None


def test_remove_dataframes(meta_partitions_files_only, store_session):
    mp = meta_partitions_files_only[0].load_dataframes(store=store_session)
    assert mp.data is not None
    mp = mp.remove_dataframes()
    assert mp.data is None


def test_load_dataframes_selective(meta_partitions_files_only, store_session):
    expected_df = pd.DataFrame(
        OrderedDict(
            [("P", [1]), ("L", [1]), ("TARGET", [1]), ("DATE", [date(2010, 1, 1)])]
        )
    )
    mp = meta_partitions_files_only[0]
    assert mp.file is not None
    assert mp.data is not None
    mp = meta_partitions_files_only[0].load_dataframes(store=store_session)

    data = mp.data

    pdt.assert_frame_equal(data, expected_df, check_dtype=False)


def test_load_dataframes_columns_projection(
    meta_partitions_evaluation_files_only, store_session
):
    expected_df = pd.DataFrame(OrderedDict([("P", [1]), ("L", [1]), ("HORIZON", [1])]))
    mp = meta_partitions_evaluation_files_only[0]
    assert mp.file is not None
    assert mp.data is not None
    mp = meta_partitions_evaluation_files_only[0].load_dataframes(
        store=store_session, columns=["P", "L", "HORIZON"]
    )

    data = mp.data

    pdt.assert_frame_equal(data, expected_df, check_dtype=False)


def test_load_dataframes_columns_raises_missing(
    meta_partitions_evaluation_files_only, store_session
):
    mp = meta_partitions_evaluation_files_only[0]
    assert mp.file is not None
    assert mp.data is not None
    with pytest.raises(ValueError) as e:
        meta_partitions_evaluation_files_only[0].load_dataframes(
            store=store_session, columns=["P", "L", "HORIZON", "foo", "bar"]
        )
    assert str(e.value) == "Columns cannot be found in stored dataframe: bar, foo"


def test_from_dict():
    df = pd.DataFrame({"a": [1]})
    dct = {"data": df, "label": "test_label"}
    meta_partition = MetaPartition.from_dict(dct)

    pdt.assert_frame_equal(meta_partition.data, df)
    assert meta_partition.metadata_version == DEFAULT_METADATA_VERSION


def test_eq():
    df = pd.DataFrame({"a": [1]})
    df_same = pd.DataFrame({"a": [1]})
    df_other = pd.DataFrame({"a": [2]})
    df_diff_col = pd.DataFrame({"b": [1]})
    df_diff_type = pd.DataFrame({"b": [1.0]})

    meta_partition = MetaPartition.from_dict({"label": "test_label", "data": df})
    assert meta_partition == meta_partition

    meta_partition_same = MetaPartition.from_dict(
        {"label": "test_label", "data": df_same}
    )
    assert meta_partition == meta_partition_same

    meta_partition_diff_label = MetaPartition.from_dict(
        {"label": "another_label", "data": df}
    )
    assert meta_partition != meta_partition_diff_label
    assert meta_partition_diff_label != meta_partition

    meta_partition_diff_files = MetaPartition.from_dict(
        {"label": "another_label", "data": df, "file": "something"}
    )
    assert meta_partition != meta_partition_diff_files
    assert meta_partition_diff_files != meta_partition

    meta_partition_diff_col = MetaPartition.from_dict(
        {"label": "test_label", "data": df_diff_col}
    )
    assert meta_partition != meta_partition_diff_col
    assert meta_partition_diff_col != meta_partition

    meta_partition_diff_type = MetaPartition.from_dict(
        {"label": "test_label", "data": df_diff_type}
    )
    assert meta_partition != meta_partition_diff_type
    assert meta_partition_diff_type != meta_partition

    meta_partition_diff_metadata = MetaPartition.from_dict(
        {"label": "test_label", "data": df_diff_type}
    )
    assert meta_partition != meta_partition_diff_metadata
    assert meta_partition_diff_metadata != meta_partition

    meta_partition_different_df = MetaPartition.from_dict(
        {"label": "test_label", "data": df_other}
    )
    assert not meta_partition == meta_partition_different_df

    meta_partition_empty_data = MetaPartition.from_dict(
        {"label": "test_label", "data": None}
    )
    assert meta_partition_empty_data == meta_partition_empty_data

    assert not meta_partition == "abc"


def test_add_nested_to_plain():
    mp = MetaPartition(
        label="label_1",
        file="file",
        data=pd.DataFrame({"test": [1, 2, 3]}),
        indices={"test": [1, 2, 3]},
    )

    to_nest = [
        MetaPartition(
            label="label_2",
            data=pd.DataFrame({"test": [4, 5, 6]}),
            indices={"test": [4, 5, 6]},
        ),
        MetaPartition(
            label="label_22",
            data=pd.DataFrame({"test": [4, 5, 6]}),
            indices={"test": [4, 5, 6]},
        ),
    ]
    mp_nested = to_nest[0].add_metapartition(to_nest[1])

    mp_add_nested = mp.add_metapartition(mp_nested)
    mp_iter = mp.add_metapartition(to_nest[0]).add_metapartition(to_nest[1])

    assert mp_add_nested == mp_iter


def test_add_nested_to_nested():
    mps1 = [
        MetaPartition(
            label="label_1",
            file="file",
            data=pd.DataFrame({"test": [1, 2, 3]}),
            indices={"test": [1, 2, 3]},
        ),
        MetaPartition(
            label="label_33",
            file="file",
            data=pd.DataFrame({"test": [1, 2, 3]}),
            indices={"test": [1, 2, 3]},
        ),
    ]

    mpn_1 = mps1[0].add_metapartition(mps1[1])

    mps2 = [
        MetaPartition(
            label="label_2",
            data=pd.DataFrame({"test": [4, 5, 6]}),
            indices={"test": [4, 5, 6]},
        ),
        MetaPartition(
            label="label_22",
            data=pd.DataFrame({"test": [4, 5, 6]}),
            indices={"test": [4, 5, 6]},
        ),
    ]
    mpn_2 = mps2[0].add_metapartition(mps2[1])

    mp_nested_merge = mpn_1.add_metapartition(mpn_2)

    mp_iter = mps1.pop()
    for mp_ in [*mps1, *mps2]:
        mp_iter = mp_iter.add_metapartition(mp_)

    assert mp_nested_merge == mp_iter


def test_eq_nested():
    mp_1 = MetaPartition(
        label="label_1",
        file="file",
        data=pd.DataFrame({"test": [1, 2, 3]}),
        indices={"test": [1, 2, 3]},
    )

    mp_2 = MetaPartition(
        label="label_2",
        data=pd.DataFrame({"test": [4, 5, 6]}),
        indices={"test": [4, 5, 6]},
    )

    mp = mp_1.add_metapartition(mp_2)

    assert mp == mp
    assert mp != mp_2
    assert mp_2 != mp

    mp_other = MetaPartition(label="label_3", data=pd.DataFrame({"test": [4, 5, 6]}))
    mp_other = mp_1.add_metapartition(mp_other)
    assert mp != mp_other
    assert mp_other != mp


def test_nested_incompatible_meta():
    mp = MetaPartition(
        label="label_1",
        data=pd.DataFrame({"test": np.array([1, 2, 3], dtype=np.int8)}),
        metadata_version=4,
    )

    mp_2 = MetaPartition(
        label="label_2",
        data=pd.DataFrame({"test": np.array([4, 5, 6], dtype=np.float64)}),
        metadata_version=4,
    )
    with pytest.raises(ValueError):
        mp.add_metapartition(mp_2)


def test_unique_label():
    label_list = ["first_0", "first_1"]

    assert _unique_label(label_list) == "first"

    label_list = ["test_0", "test_1"]

    assert _unique_label(label_list) == "test"

    label_list = ["test_0", "test_1", "test_2"]

    assert _unique_label(label_list) == "test"

    label_list = ["something", "else"]

    assert _unique_label(label_list) == "something_else"


def test_merge_indices():
    indices = [
        MetaPartition(
            label="label1",
            indices={"location": {"Loc1": ["label1"], "Loc2": ["label1"]}},
        ),
        MetaPartition(
            label="label2",
            indices={
                "location": {"Loc3": ["label2"], "Loc2": ["label2"]},
                "product": {"Product1": ["label2"], "Product2": ["label2"]},
            },
        ),
    ]
    result = MetaPartition.merge_indices(indices)
    expected = {
        "location": ExplicitSecondaryIndex(
            "location",
            {"Loc1": ["label1"], "Loc2": ["label1", "label2"], "Loc3": ["label2"]},
        ),
        "product": ExplicitSecondaryIndex(
            "product", {"Product1": ["label2"], "Product2": ["label2"]}
        ),
    }
    assert result == expected


def test_build_indices():
    columns = ["location", "product"]
    df = pd.DataFrame(
        OrderedDict(
            [("location", ["Loc1", "Loc2"]), ("product", ["Product1", "Product2"])]
        )
    )
    mp = MetaPartition(label="partition_label", data=df)
    result_mp = mp.build_indices(columns)
    result = result_mp.indices
    loc_index = ExplicitSecondaryIndex(
        "location",
        {"Loc1": ["partition_label"], "Loc2": ["partition_label"]},
        dtype=pa.large_string(),
    )
    prod_index = ExplicitSecondaryIndex(
        "product",
        {"Product1": ["partition_label"], "Product2": ["partition_label"]},
        dtype=pa.large_string(),
    )

    assert result["location"] == loc_index
    assert result["product"] == prod_index


def test_add_metapartition():
    mp = MetaPartition(
        label="label_1",
        data=pd.DataFrame({"test": [1, 2, 3]}),
        indices={"test": [1, 2, 3]},
    )

    mp_2 = MetaPartition(
        label="label_2",
        data=pd.DataFrame({"test": [4, 5, 6]}),
        indices={"test": [4, 5, 6]},
    )

    new_mp = mp.add_metapartition(mp_2)

    # Cannot access single object attributes
    with pytest.raises(AttributeError):
        new_mp.indices
    with pytest.raises(AttributeError):
        new_mp.label
    with pytest.raises(AttributeError):
        new_mp.data
    with pytest.raises(AttributeError):
        new_mp.file
    with pytest.raises(AttributeError):
        new_mp.indices
    with pytest.raises(AttributeError):
        new_mp.indices

    partition_list = new_mp.metapartitions

    assert len(partition_list) == 2

    first_mp = partition_list[0]
    assert first_mp["label"] == "label_1"
    assert first_mp["indices"] == {"test": [1, 2, 3]}

    first_mp = partition_list[1]
    assert first_mp["label"] == "label_2"
    assert first_mp["indices"] == {"test": [4, 5, 6]}

    # This tests whether it is possible to add to an already nested MetaPartition
    mp_3 = MetaPartition(
        label="label_3",
        data=pd.DataFrame({"test": [7, 8, 9]}),
        indices={"test": [7, 8, 9]},
    )
    new_mp = new_mp.add_metapartition(mp_3)

    partition_list = new_mp.metapartitions

    assert len(partition_list) == 3

    first_mp = partition_list[0]
    assert first_mp["label"] == "label_1"
    assert first_mp["indices"] == {"test": [1, 2, 3]}

    first_mp = partition_list[1]
    assert first_mp["label"] == "label_2"
    assert first_mp["indices"] == {"test": [4, 5, 6]}

    first_mp = partition_list[2]
    assert first_mp["label"] == "label_3"
    assert first_mp["indices"] == {"test": [7, 8, 9]}


def test_to_dict(metadata_version):
    df = pd.DataFrame({"A": [1]})
    schema = make_meta(df, origin="test")
    mp = MetaPartition(
        label="label_1",
        file="file",
        data=df,
        indices={"test": [1, 2, 3]},
        metadata_version=metadata_version,
        schema=schema,
    )
    mp_dct = mp.to_dict()
    assert mp_dct == {
        "label": "label_1",
        "data": df,
        "file": "file",
        "indices": {"test": [1, 2, 3]},
        "metadata_version": metadata_version,
        "schema": schema,
        "partition_keys": [],
        "logical_conjunction": None,
        "table_name": SINGLE_TABLE,
    }


def test_add_metapartition_duplicate_labels():
    mp = MetaPartition(label="label")

    mp_2 = MetaPartition(label="label")
    with pytest.raises(RuntimeError):
        mp.add_metapartition(mp_2)


def test_copy():
    mp = MetaPartition(
        label="label_1", file="file", data=pd.DataFrame(), indices={"test": [1, 2, 3]}
    )
    new_mp = mp.copy()

    # Check if the copy is identical
    assert new_mp == mp
    # ... but not the same object
    assert id(new_mp) != id(mp)

    new_mp = mp.copy(file="new_file")
    assert id(new_mp) != id(mp)
    assert new_mp.file == "new_file"

    new_mp = mp.copy(indices={"new": [1, 2, 3]})
    assert id(new_mp) != id(mp)
    assert new_mp.indices == {"new": [1, 2, 3]}


def test_nested_copy():
    mp = MetaPartition(
        label="label_1",
        file="file",
        data=pd.DataFrame({"test": [1, 2, 3]}),
        indices={"test": {1: "label_1", 2: "label_2", 3: "label_3"}},
    )

    mp_2 = MetaPartition(
        label="label_2",
        data=pd.DataFrame({"test": [4, 5, 6]}),
        indices={"test": [4, 5, 6]},
    )
    mp = mp.add_metapartition(mp_2)
    assert len(mp.metapartitions) == 2
    new_mp = mp.copy()

    # Check if the copy is identical
    assert len(new_mp.metapartitions) == len(mp.metapartitions)
    assert new_mp == mp
    # ... but not the same object
    assert id(new_mp) != id(mp)


def test_partition_on_one_level():
    original_df = pd.DataFrame({"test": [1, 2, 3], "some_values": [1, 2, 3]})
    mp = MetaPartition(
        label="label_1", file="file", data=original_df, metadata_version=4
    )

    new_mp = mp.partition_on(["test"])

    assert len(new_mp.metapartitions) == 3

    labels = set()
    for mp in new_mp:
        labels.add(mp.label)
        assert len(mp.data) == 1
        assert mp.data is not None
        df = mp.data
        assert df._mgr.is_view

        # try to be agnostic about the order
        assert len(df) == 1
        assert "test" not in df
    expected_labels = {"test=1/label_1", "test=2/label_1", "test=3/label_1"}
    assert labels == expected_labels


def test_partition_on_one_level_ts():
    original_df = pd.DataFrame(
        {
            "test": [
                pd.Timestamp("2001-01-01"),
                pd.Timestamp("2001-01-02"),
                pd.Timestamp("2001-01-03"),
            ],
            "some_values": [1, 2, 3],
        }
    )
    mp = MetaPartition(
        label="label_1", file="file", data=original_df, metadata_version=4
    )

    new_mp = mp.partition_on(["test"])

    assert len(new_mp.metapartitions) == 3

    labels = set()
    for mp in new_mp:
        labels.add(mp.label)
        assert len(mp.data) == 1
        assert mp.data is not None
        df = mp.data
        assert df._mgr.is_view

        # try to be agnostic about the order
        assert len(df) == 1
        assert "test" not in df
    expected_labels = {
        "test=2001-01-01%2000%3A00%3A00/label_1",
        "test=2001-01-02%2000%3A00%3A00/label_1",
        "test=2001-01-03%2000%3A00%3A00/label_1",
    }
    assert labels == expected_labels


def test_partition_on_roundtrip(store):
    original_df = pd.DataFrame(
        OrderedDict([("test", [1, 2, 3]), ("some_values", [1, 2, 3])])
    )
    mp = MetaPartition(label="label_1", data=original_df, metadata_version=4)

    new_mp = mp.partition_on(["test"])
    new_mp = new_mp.store_dataframes(store=store, dataset_uuid="some_uuid")
    store_schema_metadata(new_mp.schema, "some_uuid", store)
    # Test immediately after dropping and later once with new metapartition to check table meta reloading
    new_mp = new_mp.load_dataframes(store=store)
    assert len(new_mp.metapartitions) == 3
    dfs = []
    for internal_mp in new_mp:
        dfs.append(internal_mp.data)
    actual_df = pd.concat(dfs).sort_values(by="test").reset_index(drop=True)
    pdt.assert_frame_equal(original_df, actual_df)

    for i in range(1, 4):
        # Check with fresh metapartitions
        new_mp = MetaPartition(
            label=f"test={i}/label_1",
            file=f"some_uuid/table/test={i}/label_1.parquet",
            metadata_version=4,
        )
        new_mp = new_mp.load_dataframes(store=store)

        actual_df = new_mp.data

        expected_df = pd.DataFrame(OrderedDict([("test", [i]), ("some_values", [i])]))
        pdt.assert_frame_equal(expected_df, actual_df)


@pytest.mark.parametrize("empty", [True, False])
def test_partition_on_raises_no_cols_left(empty):
    original_df = pd.DataFrame({"test": [1, 2, 3]})
    if empty:
        original_df = original_df.loc[[]]
    mp = MetaPartition(
        label="label_1", file="file", data=original_df, metadata_version=4
    )
    with pytest.raises(ValueError) as e:
        mp.partition_on(["test"])
    assert str(e.value) == "No data left to save outside partition columns"


@pytest.mark.parametrize("empty", [True, False])
def test_partition_on_raises_pocols_missing(empty):
    original_df = pd.DataFrame({"test": [1, 2, 3]})
    if empty:
        original_df = original_df.loc[[]]
    mp = MetaPartition(
        label="label_1", file="file", data=original_df, metadata_version=4
    )
    with pytest.raises(ValueError) as e:
        mp.partition_on(["test", "foo", "bar"])
    assert str(e.value) == "Partition column(s) missing: bar, foo"


def test_partition_urlencode():
    original_df = pd.DataFrame({"ÖŒå": [1, 2, 3], "some_values": [1, 2, 3]})
    mp = MetaPartition(label="label_1", data=original_df, metadata_version=4)

    new_mp = mp.partition_on(["ÖŒå"])

    assert len(new_mp.metapartitions) == 3

    labels = set()
    for mp in new_mp:
        labels.add(mp.label)
        assert len(mp.data) == 1
        assert mp.data is not None
        df = mp.data
        assert df._mgr.is_view

        # try to be agnostic about the order
        assert len(df) == 1
        assert "ÖŒå" not in df
    expected_labels = {
        "%C3%96%C5%92%C3%A5=1/label_1",
        "%C3%96%C5%92%C3%A5=2/label_1",
        "%C3%96%C5%92%C3%A5=3/label_1",
    }
    assert labels == expected_labels


def test_partition_two_level():
    original_df = pd.DataFrame(
        {
            "level1": [1, 2, 3, 1, 2, 3],
            "level2": [1, 1, 1, 2, 2, 2],
            "no_index_col": np.arange(0, 6),
        }
    )
    mp = MetaPartition(
        label="label_1", file="file", data=original_df, metadata_version=4
    )

    new_mp = mp.partition_on(["level1", "level2"])
    assert len(new_mp.metapartitions) == 6

    labels = []
    for mp in new_mp:
        labels.append(mp.label)
        assert len(mp.data) == 1
        assert mp.data is not None
        df = mp.data
        assert df._mgr.is_view

        # try to be agnostic about the order
        assert len(df) == 1
        assert "level1" not in df
        assert "level2" not in df
        assert "no_index_col" in df
    expected_labels = [
        "level1=1/level2=1/label_1",
        "level1=1/level2=2/label_1",
        "level1=2/level2=1/label_1",
        "level1=2/level2=2/label_1",
        "level1=3/level2=1/label_1",
        "level1=3/level2=2/label_1",
    ]
    assert sorted(labels) == sorted(expected_labels)


def test_partition_on_nested():
    original_df = pd.DataFrame(
        {
            "level1": [1, 2, 3, 1, 2, 3],
            "level2": [1, 1, 1, 2, 2, 2],
            "no_index_col": np.arange(0, 6),
        }
    )
    mp = MetaPartition(
        label="label_1", file="file", data=original_df, metadata_version=4
    )
    mp2 = MetaPartition(
        label="label_2", file="file", data=original_df, metadata_version=4
    )
    mp = mp.add_metapartition(mp2)
    new_mp = mp.partition_on(["level1", "level2"])
    assert len(new_mp.metapartitions) == 12

    labels = []
    for mp in new_mp:
        labels.append(mp.label)
        assert len(mp.data) == 1
        assert mp.data is not None
        df = mp.data
        assert df._mgr.is_view

        # try to be agnostic about the order
        assert len(df) == 1
        assert "level1" not in df
        assert "level2" not in df
        assert "no_index_col" in df
    expected_labels = [
        "level1=1/level2=1/label_1",
        "level1=1/level2=2/label_1",
        "level1=2/level2=1/label_1",
        "level1=2/level2=2/label_1",
        "level1=3/level2=1/label_1",
        "level1=3/level2=2/label_1",
        "level1=1/level2=1/label_2",
        "level1=1/level2=2/label_2",
        "level1=2/level2=1/label_2",
        "level1=2/level2=2/label_2",
        "level1=3/level2=1/label_2",
        "level1=3/level2=2/label_2",
    ]
    assert sorted(labels) == sorted(expected_labels)


def test_partition_on_stable_order():
    """Assert that the partition_on algo is stable wrt to row ordering."""
    unique_values = 3
    total_values = 20
    random_index = np.repeat(
        np.arange(unique_values), int(np.ceil(total_values / unique_values))
    )[:total_values]
    np.random.shuffle(random_index)
    df = pd.DataFrame(
        {"partition_key": random_index, "sorted_col": range(total_values)}
    )
    mp = MetaPartition(label="label_1", data=df, metadata_version=4)
    new_mp = mp.partition_on("partition_key")
    for sub_mp in new_mp:
        sub_df = sub_mp.data
        assert sub_df.sorted_col.is_monotonic_increasing


def test_table_meta(store):
    mp = MetaPartition(
        label="label_1",
        data=pd.DataFrame(
            {
                "i32": np.array([1, 2, 3, 1, 2, 3], dtype="int32"),
                "float": np.array([1, 1, 1, 2, 2, 2], dtype="float64"),
            }
        ),
        metadata_version=4,
    )

    assert mp.schema is not None
    expected_meta = make_meta(
        pd.DataFrame(
            {"i32": np.array([], dtype="int32"), "float": np.array([], dtype="float64")}
        ),
        origin="1",
    )
    actual_meta = mp.schema
    assert actual_meta == expected_meta

    mp = mp.store_dataframes(store, "dataset_uuid")

    actual_meta = mp.schema
    assert actual_meta == expected_meta


def test_partition_on_explicit_index():
    original_df = pd.DataFrame(
        {
            "level1": [1, 2, 1, 2, 1, 2],
            "level2": [1, 1, 1, 2, 2, 2],
            "explicit_index_col": np.arange(0, 6),
        }
    )
    mp = MetaPartition(
        label="label_1",
        file="file",
        data=original_df,
        indices={
            "explicit_index_col": {value: ["label_1"] for value in np.arange(0, 6)}
        },
        metadata_version=4,
    )
    new_mp = mp.partition_on(["level1", "level2"])
    assert len(new_mp) == 4

    expected_indices = {
        "explicit_index_col": ExplicitSecondaryIndex(
            "explicit_index_col",
            {0: ["level1=1/level2=1/label_1"], 2: ["level1=1/level2=1/label_1"]},
        )
    }
    assert expected_indices == new_mp["level1=1/level2=1/label_1"].indices

    expected_indices = {
        "explicit_index_col": ExplicitSecondaryIndex(
            "explicit_index_col", {4: ["level1=1/level2=2/label_1"]}
        )
    }
    assert expected_indices == new_mp["level1=1/level2=2/label_1"].indices

    expected_indices = {
        "explicit_index_col": ExplicitSecondaryIndex(
            "explicit_index_col", {1: ["level1=2/level2=1/label_1"]}
        )
    }
    assert expected_indices == new_mp["level1=2/level2=1/label_1"].indices

    expected_indices = {
        "explicit_index_col": ExplicitSecondaryIndex(
            "explicit_index_col",
            {3: ["level1=2/level2=2/label_1"], 5: ["level1=2/level2=2/label_1"]},
        )
    }
    assert expected_indices == new_mp["level1=2/level2=2/label_1"].indices


def test_reconstruct_index_duplicates(store):
    ser = ParquetSerializer()
    df = pd.DataFrame({"index_col": [1, 1], "column": list("ab")})

    label = "dontcare"
    key_prefix = f"uuid/table/index_col=2/{label}"
    key = ser.store(store, key_prefix, df)

    schema = make_meta(df, origin="1", partition_keys="index_col")
    store_schema_metadata(schema, "uuid", store)

    mp = MetaPartition(
        label="dontcare",
        file=key,
        metadata_version=4,
        schema=schema,
        partition_keys=["index_col"],
    )
    mp = mp.load_dataframes(store)
    df_actual = mp.data
    df_expected = pd.DataFrame(
        OrderedDict([("index_col", [2, 2]), ("column", list("ab"))])
    )
    pdt.assert_frame_equal(df_actual, df_expected)


def test_reconstruct_index_categories(store):
    ser = ParquetSerializer()
    df = pd.DataFrame(
        {"index_col": [1, 1], "second_index_col": [2, 2], "column": list("ab")}
    )

    label = "dontcare"
    key_prefix = f"uuid/table/index_col=2/second_index_col=2/{label}"
    key = ser.store(store, key_prefix, df)

    schema = make_meta(df, origin="1", partition_keys="index_col")
    store_schema_metadata(schema, "uuid", store)

    mp = MetaPartition(
        label="index_col=2/dontcare",
        file=key,
        metadata_version=4,
        schema=schema,
        partition_keys=["index_col", "second_index_col"],
    )
    categories = ["second_index_col", "index_col"]
    mp = mp.load_dataframes(store, categoricals=categories)
    df_actual = mp.data
    df_expected = pd.DataFrame(
        OrderedDict(
            [
                ("index_col", [2, 2]),
                ("second_index_col", [2, 2]),
                ("column", list("ab")),
            ]
        )
    )
    df_expected = df_expected.astype(dict.fromkeys(categories, "category"))
    pdt.assert_frame_equal(df_actual, df_expected)


@pytest.mark.parametrize("categoricals", [True, False])
def test_reconstruct_index_empty_df(store, categoricals):
    ser = ParquetSerializer()
    df = pd.DataFrame({"index_col": [1, 1], "column": list("ab")})
    df = df[0:0]

    label = "dontcare"
    key_prefix = f"uuid/table/index_col=2/{label}"
    key = ser.store(store, key_prefix, df)

    schema = make_meta(df, origin="1", partition_keys="index_col")
    store_schema_metadata(schema, "uuid", store)

    mp = MetaPartition(
        label="index_col=2/dontcare",
        file=key,
        metadata_version=4,
        schema=schema,
        partition_keys=["index_col"],
    )
    categoricals = None
    if categoricals:
        categoricals = ["index_col"]
    mp = mp.load_dataframes(store, categoricals=categoricals)
    df_actual = mp.data
    df_expected = pd.DataFrame(
        OrderedDict([("index_col", [2, 2]), ("column", list("ab"))])
    )
    df_expected = df_expected[0:0]
    if categoricals:
        df_expected = df_expected.astype({"index_col": "category"})
    pdt.assert_frame_equal(df_actual, df_expected)


@pytest.mark.parametrize("dates_as_object", [True, False])
def test_reconstruct_date_index(store, metadata_version, dates_as_object):
    ser = ParquetSerializer()
    # If the parquet file does include the primary index col, still use the reconstructed index and ignore the content of the file
    df = pd.DataFrame(
        {"index_col": [date(2018, 6, 1), date(2018, 6, 1)], "column": list("ab")}
    )

    label = "dontcare"
    key_prefix = f"uuid/table/index_col=2018-06-02/{label}"
    key = ser.store(store, key_prefix, df)

    schema = make_meta(df, origin="1", partition_keys="index_col")
    store_schema_metadata(schema, "uuid", store)

    mp = MetaPartition(
        label="dontcare",
        file=key,
        metadata_version=metadata_version,
        schema=schema,
        partition_keys=["index_col"],
    )

    mp = mp.load_dataframes(store, dates_as_object=dates_as_object)
    df_actual = mp.data
    if dates_as_object:
        index_col = [date(2018, 6, 2), date(2018, 6, 2)]
    else:
        index_col = pd.Series(
            [pd.Timestamp("2018-06-02"), pd.Timestamp("2018-06-02")],
            dtype="datetime64[ns]",
        )
    df_expected = pd.DataFrame(
        OrderedDict(
            [
                ("index_col", index_col),
                ("column", list("ab")),
            ]
        )
    )
    pdt.assert_frame_equal(df_actual, df_expected)


def test_iter_empty_metapartition():
    for _ in MetaPartition(None):
        raise AssertionError(
            "Iterating over an empty MetaPartition should stop immediately"
        )


def test_concat_metapartition(df_all_types):
    mp1 = MetaPartition(label="first", data=df_all_types, metadata_version=4)
    mp2 = MetaPartition(label="second", data=df_all_types, metadata_version=4)

    new_mp = MetaPartition.concat_metapartitions([mp1, mp2])

    # what the label actually is, doesn't matter so much
    assert new_mp.label is not None
    df_expected = pd.concat([df_all_types, df_all_types])
    df_actual = new_mp.data
    pdt.assert_frame_equal(df_actual, df_expected)


def test_concat_metapartition_wrong_types(df_all_types):
    mp1 = MetaPartition(label="first", data=df_all_types, metadata_version=4)
    df_corrupt = df_all_types.copy()
    df_corrupt["int8"] = "NoInteger"
    mp2 = MetaPartition(label="second", data=df_corrupt, metadata_version=4)

    with pytest.raises(ValueError, match="Schema violation"):
        MetaPartition.concat_metapartitions([mp1, mp2])


def test_concat_metapartition_partitioned(df_all_types):
    mp1 = MetaPartition(
        label="int8=1/1234",
        data=df_all_types,
        metadata_version=4,
        partition_keys=["int8"],
    )
    mp2 = MetaPartition(
        label="int8=1/4321",
        data=df_all_types,
        metadata_version=4,
        partition_keys=["int8"],
    )

    new_mp = MetaPartition.concat_metapartitions([mp1, mp2])

    df_expected = pd.concat([df_all_types, df_all_types])
    df_actual = new_mp.data
    pdt.assert_frame_equal(df_actual, df_expected)

    assert new_mp.partition_keys == ["int8"]


def test_concat_metapartition_different_partitioning(df_all_types):
    mp1 = MetaPartition(
        label="int8=1/1234",
        data=df_all_types,
        metadata_version=4,
        partition_keys=["int8"],
    )
    mp2 = MetaPartition(
        label="float8=1.0/4321",
        data=df_all_types,
        metadata_version=4,
        partition_keys=["float8"],
    )

    with pytest.raises(ValueError, match="Schema violation"):
        MetaPartition.concat_metapartitions([mp1, mp2])


def test_concat_metapartition_categoricals(df_all_types):
    mp1 = MetaPartition(
        label="first",
        data=pd.DataFrame({"a": [0, 0], "b": ["a", "a"]}, dtype="category"),
        metadata_version=4,
        partition_keys=["a"],
    )
    mp2 = MetaPartition(
        label="second",
        data=pd.DataFrame({"a": [1, 1], "b": ["a", "b"]}, dtype="category"),
        metadata_version=4,
        partition_keys=["a"],
    )

    new_mp = MetaPartition.concat_metapartitions([mp1, mp2])

    assert new_mp.table_name == "table"
    assert isinstance(new_mp.data["b"].dtype, pd.CategoricalDtype)


# We can't partition on null columns (gh-262)
@pytest.mark.parametrize(
    "col", sorted(set(get_dataframe_not_nested().columns) - {"null"})
)
def test_partition_on_scalar_intermediate(df_not_nested, col):
    """Test against a bug where grouping leaves a scalar value."""
    assert len(df_not_nested) == 1
    mp = MetaPartition(label="somelabel", data=df_not_nested, metadata_version=4)
    new_mp = mp.partition_on(col)
    assert len(new_mp) == 1


def test_partition_on_with_primary_index_invalid(df_not_nested):
    mp = MetaPartition(
        label="pkey=1/pkey2=2/base_label",
        data=df_not_nested,
        partition_keys=["pkey", "pkey2"],
        metadata_version=4,
    )
    with pytest.raises(ValueError, match="Incompatible"):
        mp.partition_on("int64")

    with pytest.raises(ValueError, match="Incompatible"):
        mp.partition_on(["int64", "pkey"])

    with pytest.raises(ValueError, match="Incompatible"):
        mp.partition_on(["pkey", "int64"])

    with pytest.raises(ValueError, match="Incompatible"):
        mp.partition_on(["pkey2", "pkey1", "int64"])

    mp.partition_on(["pkey", "pkey2"])
    mp.partition_on(["pkey", "pkey2", "int64"])


def test_partition_on_with_primary_index(df_not_nested):
    mp = MetaPartition(
        label="pkey=1/base_label",
        data=df_not_nested,
        partition_keys=["pkey"],
        metadata_version=4,
    )
    new = mp.partition_on(["pkey", "int64"])

    split_label = new.label.split("/")

    assert len(split_label) == 3
    assert split_label[0] == "pkey=1"
    assert split_label[1] == "int64=1"
    assert split_label[2] == "base_label"

    assert mp == mp.partition_on(["pkey"])


@pytest.mark.parametrize(
    "labels, flat_labels",
    [
        ([], []),
        (["a"], ["a"]),
        (["a", "b", "c"], ["a", "b", "c"]),
        (["a", None], ["a"]),
        (["a", ["b", "c"]], ["a", "b", "c"]),
        (["a", None, ["b", "c", None]], ["a", "b", "c"]),
    ],
)
def test_partition_label_helper(labels, flat_labels):
    mps = []
    for lbl in labels:
        if isinstance(lbl, list):
            mp = MetaPartition(lbl[0])
            for nested_lbl in lbl[1:]:
                mp = mp.add_metapartition(MetaPartition(label=nested_lbl))
            mps.append(mp)
        else:
            mps.append(MetaPartition(label=lbl))

    assert set(partition_labels_from_mps(mps)) == set(flat_labels)


def test_column_string_cast(df_all_types, store, metadata_version):
    original_columns = df_all_types.columns.copy()
    df_all_types.columns = df_all_types.columns.str.encode("utf-8")
    ser = ParquetSerializer()
    key = ser.store(store, "uuid/table/something", df_all_types)
    mp = MetaPartition(
        label="something",
        file=key,
        schema=make_meta(df_all_types, origin="table"),
        metadata_version=metadata_version,
    )
    mp = mp.load_dataframes(store)
    df = mp.data
    assert all(original_columns == df.columns)


def test_partition_on_valid_schemas():
    """Ensure that partitioning is possible even if the output schemas of the
    sub partitions may be different."""
    df = pd.DataFrame({"partition_col": [0, 1], "values": [None, "str"]})
    mp = MetaPartition(label="base_label", data=df, metadata_version=4)
    mp = mp.partition_on(["partition_col"])
    assert len(mp) == 2
    expected_meta = make_meta(df, origin="1", partition_keys="partition_col")
    assert mp.schema == expected_meta


def test_input_to_metaframes_empty():
    mp = parse_input_to_metapartition(obj=[None])
    assert mp == MetaPartition(label=None)
    mp = parse_input_to_metapartition(obj=[])
    assert mp == MetaPartition(label=None)


def test_input_to_metaframes_simple():
    df_input = pd.DataFrame({"A": [1]})
    mp = parse_input_to_metapartition(obj=df_input)

    assert isinstance(mp, MetaPartition)
    assert len(mp.data) == 1
    assert mp.file is None

    df = mp.data
    pdt.assert_frame_equal(df, df_input)

    assert isinstance(mp.label, str)


def test_parse_nested_input_schema_compatible_but_different():
    # Ensure that input can be parsed even though the schemas are not identical but compatible
    df_input = [[pd.DataFrame({"A": [None]}), pd.DataFrame({"A": ["str"]})]]
    mp = parse_input_to_metapartition(df_input, metadata_version=4)
    expected_schema = make_meta(pd.DataFrame({"A": ["str"]}), origin="expected")
    assert mp.schema == expected_schema


def test_get_parquet_metadata(store):
    df = pd.DataFrame({"P": np.arange(0, 10), "L": np.arange(0, 10)})
    mp = MetaPartition(label="test_label", data=df)
    meta_partition = mp.store_dataframes(store=store, dataset_uuid="dataset_uuid")

    actual = meta_partition.get_parquet_metadata(store=store)
    actual.drop(labels="serialized_size", axis=1, inplace=True)
    actual.drop(labels="row_group_compressed_size", axis=1, inplace=True)
    actual.drop(labels="row_group_uncompressed_size", axis=1, inplace=True)

    expected = pd.DataFrame(
        {
            "partition_label": ["test_label"],
            "row_group_id": 0,
            "number_rows_total": 10,
            "number_row_groups": 1,
            "number_rows_per_row_group": 10,
        }
    )
    pd.testing.assert_frame_equal(actual, expected)


def test_get_parquet_metadata_empty_df(store):
    df = pd.DataFrame()
    mp = MetaPartition(label="test_label", data=df)
    meta_partition = mp.store_dataframes(store=store, dataset_uuid="dataset_uuid")

    actual = meta_partition.get_parquet_metadata(store=store)
    actual = actual.drop(
        columns=[
            "serialized_size",
            "row_group_compressed_size",
            "row_group_uncompressed_size",
        ]
    )

    expected = pd.DataFrame(
        {
            "partition_label": ["test_label"],
            "row_group_id": 0,
            "number_rows_total": 0,
            "number_row_groups": 1,
            "number_rows_per_row_group": 0,
        }
    )

    pd.testing.assert_frame_equal(actual, expected)


def test_get_parquet_metadata_row_group_size(store):
    df = pd.DataFrame({"P": np.arange(0, 10), "L": np.arange(0, 10)})
    mp = MetaPartition(label="test_label", data=df)
    ps = ParquetSerializer(chunk_size=5)

    meta_partition = mp.store_dataframes(
        store=store, dataset_uuid="dataset_uuid", df_serializer=ps
    )
    actual = meta_partition.get_parquet_metadata(store=store)
    actual = actual.drop(
        columns=[
            "serialized_size",
            "row_group_compressed_size",
            "row_group_uncompressed_size",
        ],
    )

    expected = pd.DataFrame(
        {
            "partition_label": ["test_label", "test_label"],
            "row_group_id": [0, 1],
            "number_rows_total": [10, 10],
            "number_row_groups": [2, 2],
            "number_rows_per_row_group": [5, 5],
        }
    )
    pd.testing.assert_frame_equal(actual, expected)


def test__reconstruct_index_columns():
    df = pd.DataFrame({"x": [0], "a": [-1], "b": [-2], "c": [-3]})
    mp = MetaPartition(label="test_label", data=df)
    df_with_index_columns = mp._reconstruct_index_columns(
        df=df[["x"]],
        key_indices=[("a", 1), ("b", 2), ("c", 3)],
        columns=["x", "c"],
        categories=None,
        date_as_object=False,
    )
    # Index columns first
    pdt.assert_frame_equal(df_with_index_columns, pd.DataFrame({"c": [3], "x": [0]}))


def test_partition_on_keeps_table_name():
    mp = MetaPartition(
        label="label_1",
        data=pd.DataFrame({"P": [1, 2, 1, 2], "L": [1, 1, 2, 2]}),
        table_name="non-default-name",
    )
    repartitioned_mp = mp.partition_on(["P"])
    assert repartitioned_mp.table_name == "non-default-name"



"""Helper module to convert user inputs into normalized forms."""

from collections.abc import Iterable

import pandas as pd
import pyarrow as pa

__all__ = (
    "converter_str",
    "converter_str_set",
    "converter_str_set_optional",
    "converter_str_tupleset",
    "converter_tuple",
    "get_str_to_python_converter",
)


def converter_str_set(obj) -> frozenset:
    """Convert input to a set of unicode strings. ``None`` will be converted to
    an empty set.

    Parameters
    ----------
    obj: Optional[Union[Iterable[str], str]]
        Object to convert.

    Returns
    -------
    obj: FrozenSet[str]
        String set.

    Raises
    ------
    TypeError
        If passed object is not string/byte-like.
    """
    result = converter_tuple(obj)
    result_set = {converter_str(x) for x in result}
    return frozenset(result_set)


def converter_str_set_optional(obj):
    """Convert input to a set of unicode strings. ``None`` will be preserved.

    Parameters
    ----------
    obj: Optional[Union[Iterable[str], str]]
        Object to convert.

    Returns
    -------
    obj: Optional[FrozenSet[str]]
        String set.

    Raises
    ------
    ValueError
        If an element in the passed object is not string/byte/like.
    """
    if obj is None:
        return None
    return converter_str_set(obj)


def converter_str_tupleset(obj: Iterable[str] | str | None) -> tuple[str, ...]:
    """Convert input to tuple of unique unicode strings. ``None`` will be
    converted to an empty set.

    The input must not contain duplicate entries.

    Parameters
    ----------
    obj
        Object to convert.

    Raises
    ------
    TypeError
        If passed object is not string/byte-like, or if ``obj`` is known to have an unstable iteration order.
    ValueError
        If passed set contains duplicates.
    """
    if isinstance(obj, dict | frozenset | set):
        raise TypeError(
            f"{obj} which has type {type(obj).__name__} has an unstable iteration order"
        )
    result = converter_tuple(obj)
    result = tuple(converter_str(x) for x in result)
    if len(set(result)) != len(result):
        raise ValueError("Tuple-set contains duplicates: {}".format(", ".join(result)))
    return result


def converter_tuple(obj) -> tuple:
    """Convert input to a tuple. ``None`` will be converted to an empty tuple.

    Parameters
    ----------
    obj: Any
        Object to convert.

    Returns
    -------
    obj: Tuple[Any]
        Tuple.
    """
    if obj is None:
        return ()
    elif hasattr(obj, "__iter__") and not isinstance(obj, str | bytes):
        return tuple(x for x in obj)
    else:
        return (obj,)


def converter_str(obj) -> str:
    """Ensures input is a unicode string.

    Parameters
    ----------
    obj: str
        Object to convert.

    Returns
    -------
    obj: str
        String.

    Raises
    ------
    TypeError
        If passed object is not string/byte-like.
    """
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, bytes):
        return obj.decode("utf-8")
    else:
        raise TypeError(f"Object of type {type(obj).__name__} is not a string: {obj}")


def get_str_to_python_converter(pa_type):
    """Get converter to parse string into python object.

    Parameters
    ----------
    pa_type: pyarrow.DataType
        Data type.

    Returns
    -------
    converter: Callable[[str], Any]
        Converter.
    """
    if pa.types.is_boolean(pa_type):

        def var_f(x):
            if x.lower() in ("0", "f", "n", "false", "no"):
                return False
            elif x.lower() in ("1", "t", "y", "true", "yes"):
                return True
            else:
                raise ValueError(f"Cannot parse bool: {x}")

        return var_f
    elif pa.types.is_floating(pa_type):
        return float
    elif pa.types.is_integer(pa_type):
        return int
    elif pa.types.is_string(pa_type):

        def var_f(x):
            if len(x) > 1:
                for char in ('"', "'"):
                    if x.startswith(char) and x.endswith(char):
                        return x[1:-1]
            return x

        return var_f
    elif pa.types.is_timestamp(pa_type):
        return pd.Timestamp
    else:
        raise ValueError(f"Cannot handle type {pa_type}")



import difflib
import logging
import pprint
from collections.abc import Sequence
from copy import copy, deepcopy
from functools import reduce
from typing import Any

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import simplejson
from minimalkv import KeyValueStore

from plateau.core import naming
from plateau.core._compat import load_json
from plateau.core.naming import SINGLE_TABLE
from plateau.core.utils import ensure_string_type
from plateau.serialization._parquet import PARQUET_VERSION
from plateau.serialization._util import schema_metadata_bytes_to_object

_logger = logging.getLogger()

__all__ = (
    "SchemaWrapper",
    "make_meta",
    "validate_compatible",
    "empty_dataframe_from_schema",
    "normalize_type",
    "normalize_column_order",
)


class SchemaWrapper:
    """Wrapper object for pyarrow.Schema to handle forwards and backwards
    compatibility."""

    def __init__(self, schema, origin: str | set[str]):
        if isinstance(origin, str):
            origin = {origin}
        elif isinstance(origin, set):
            origin = copy(origin)
        if not all(isinstance(s, str) for s in origin):
            raise TypeError("Schema origin elements must be strings.")

        self.__schema = schema
        self.__origin = origin
        self._schema_compat()

    def with_origin(self, origin: str | set[str]) -> "SchemaWrapper":
        """Create new SchemaWrapper with given origin.

        Parameters
        ----------
        origin:
            New origin.
        """
        return SchemaWrapper(self.__schema, origin)

    def _schema_compat(self):
        # https://issues.apache.org/jira/browse/ARROW-5104
        schema = self.__schema
        if self.__schema is not None and self.__schema.pandas_metadata is not None:
            pandas_metadata = schema.pandas_metadata
            index_cols = pandas_metadata["index_columns"]
            if len(index_cols) > 1:
                raise NotImplementedError("Treatment of MultiIndex not implemented.")

            for _, col in enumerate(index_cols):
                # Range index is now serialized using start/end information. This special treatment
                # removes it from the columns which is fine
                if isinstance(col, dict):
                    pass
                # other indices are still tracked as a column
                else:
                    index_level_ix = schema.get_field_index(col)
                    # this may happen for the schema of an empty df
                    if index_level_ix >= 0:
                        schema = schema.remove(index_level_ix)

            for cmd in pandas_metadata["columns"]:
                name = cmd.get("name")
                if name is None:
                    continue

                field_name = cmd["field_name"]
                field_idx = schema.get_field_index(field_name)
                if field_idx < 0:
                    continue
                field = schema[field_idx]
                if (
                    pa.types.is_string(field.type)
                    and cmd["pandas_type"] == "unicode"
                    and cmd["numpy_type"] == "object"
                ):
                    schema = schema.remove(field_idx)
                    new_field = pa.field(
                        field.name,
                        pa.large_string(),
                        field.nullable,
                        field.metadata,
                    )
                    schema = schema.insert(field_idx, new_field)
                    cmd["pandas_type"] = "object"
                    cmd["numpy_type"] = "str"
                    cmd["metadata"] = None

            schema = schema.remove_metadata()
            md = {b"pandas": _dict_to_binary(pandas_metadata)}
            schema = schema.with_metadata(md)
            self.__schema = schema

    def internal(self):
        return self.__schema

    @property
    def origin(self) -> set[str]:
        return copy(self.__origin)

    def __repr__(self):
        return self.__schema.__repr__()

    def __eq__(self, other):
        return self.equals(other)

    def __ne__(self, other):
        return not self.equals(other)

    def __getstate__(self):
        return (_schema2bytes(self.__schema), self.__origin)

    def __setstate__(self, state):
        self.__schema = _bytes2schema(state[0])
        self.__origin = state[1]

    def __getattr__(self, attr):
        return getattr(self.__schema, attr)

    def __hash__(self):
        # see https://issues.apache.org/jira/browse/ARROW-2719
        return hash(_schema2bytes(self.__schema))

    def __getitem__(self, i):
        return self.__schema[i]

    def __len__(self):
        return len(self.__schema)

    def equals(self, other, check_metadata=False):
        if isinstance(other, SchemaWrapper):
            return self.__schema.equals(other.__schema, check_metadata)
        else:
            return self.__schema.equals(other, check_metadata)

    equals.__doc__ = pa.Schema.equals.__doc__

    def remove(self, i):
        return SchemaWrapper(self.__schema.remove(i), self.__origin)

    remove.__doc__ = pa.Schema.set.__doc__

    def remove_metadata(self):
        return SchemaWrapper(
            self.__schema.remove_metadata(),
            {s + "__no_metadata" for s in self.__origin},
        )

    remove_metadata.__doc__ = pa.Schema.remove_metadata.__doc__

    def set(self, i, field):
        return SchemaWrapper(self.__schema.set(i, field), self.__origin)

    set.__doc__ = pa.Schema.set.__doc__


def normalize_column_order(schema, partition_keys=None):
    """Normalize column order in schema.

    Columns are sorted in the following way:

    1. Partition keys (as provided by ``partition_keys``)
    2. DataFrame columns in alphabetic order
    3. Remaining fields as generated by pyarrow, mostly index columns

    Parameters
    ----------
    schema: SchemaWrapper
        Schema information for DataFrame.
    partition_keys: Union[None, List[str]]
        Partition keys used to split the dataset.

    Returns
    -------
    schema: SchemaWrapper
        Schema information for DataFrame.
    """
    if not isinstance(schema, SchemaWrapper):
        schema = SchemaWrapper(schema, "__unknown__")

    if partition_keys is None:
        partition_keys = []
    else:
        partition_keys = list(partition_keys)

    pandas_metadata = schema.pandas_metadata
    origin = schema.origin

    cols_partition = {}
    cols_payload = []
    cols_misc = []

    for cmd in pandas_metadata["columns"]:
        name = cmd.get("name")
        field_name = cmd["field_name"]
        field_idx = schema.get_field_index(field_name)

        if field_idx >= 0:
            field = schema[field_idx]
        else:
            field = None

        if name is None:
            cols_misc.append((cmd, field))
        elif name in partition_keys:
            cols_partition[name] = (cmd, field)
        else:
            cols_payload.append((name, cmd, field))

    ordered = []
    for k in partition_keys:
        if k in cols_partition:
            ordered.append(cols_partition[k])

    ordered += [(cmd, f) for _name, cmd, f in sorted(cols_payload, key=lambda x: x[0])]
    ordered += cols_misc

    pandas_metadata["columns"] = [cmd for cmd, _ in ordered]
    fields = [f for _, f in ordered if f is not None]

    metadata = schema.metadata
    metadata[b"pandas"] = _dict_to_binary(pandas_metadata)
    schema = pa.schema(fields, metadata)
    return SchemaWrapper(schema, origin)


def make_meta(obj, origin, partition_keys=None):
    """Create metadata object for DataFrame.

    .. note::
        This function can, for convenience reasons, also be applied to schema objects in which case they are just
        returned.

    .. warning::
        Information for categoricals will be stripped!

    :meth:`normalize_type` will be applied to normalize type information and :meth:`normalize_column_order` will be
    applied to to reorder column information.

    Parameters
    ----------
    obj: Union[DataFrame, Schema]
        Object to extract metadata from.
    origin: str
        Origin of the schema data, used for debugging and error reporting.
    partition_keys: Union[None, List[str]]
        Partition keys used to split the dataset.

    Returns
    -------
    schema: SchemaWrapper
        Schema information for DataFrame.
    """
    if isinstance(obj, SchemaWrapper):
        return obj
    if isinstance(obj, pa.Schema):
        return normalize_column_order(
            SchemaWrapper(obj, origin), partition_keys=partition_keys
        )

    if not isinstance(obj, pd.DataFrame):
        raise ValueError("Input must be a pyarrow schema, or a pandas dataframe")

    schema = pa.Schema.from_pandas(obj)
    pandas_metadata = schema.pandas_metadata

    # normalize types
    fields = {field.name: field.type for field in schema}
    for cmd in pandas_metadata["columns"]:
        name = cmd.get("name")
        if name is None:
            continue
        field_name = cmd["field_name"]
        field_idx = schema.get_field_index(field_name)
        field = schema[field_idx]
        (
            fields[field_name],
            cmd["pandas_type"],
            cmd["numpy_type"],
            cmd["metadata"],
        ) = normalize_type(
            field.type, cmd["pandas_type"], cmd["numpy_type"], cmd["metadata"]
        )
    metadata = schema.metadata
    metadata[b"pandas"] = _dict_to_binary(pandas_metadata)
    schema = pa.schema([pa.field(n, t) for n, t in fields.items()], metadata)
    return normalize_column_order(SchemaWrapper(schema, origin), partition_keys)


def normalize_type(
    t_pa: pa.DataType,
    t_pd: str | None,
    t_np: str | None,
    metadata: dict[str, Any] | None,
) -> tuple[pa.DataType, str | None, str | None, dict[str, Any] | None]:
    """This will normalize types as followed:

    - all signed integers (``int8``, ``int16``, ``int32``, ``int64``) will be converted to ``int64``
    - all unsigned integers (``uint8``, ``uint16``, ``uint32``, ``uint64``) will be converted to ``uint64``
    - all floats (``float32``, ``float64``) will be converted to ``float64``
    - all list value types will be normalized (e.g. ``list[int16]`` to ``list[int64]``, ``list[list[uint8]]`` to
      ``list[list[uint64]]``)
    - all dict value types will be normalized (e.g. ``dictionary<values=float32, indices=int16, ordered=0>`` to
      ``float64``)

    Parameters
    ----------
    t_pa
        pyarrow type object, e.g. ``pa.list_(pa.int8())``.
    t_pd
        pandas type identifier, e.g. ``"list[int8]"``.
    t_np
        numpy type identifier, e.g. ``"object"``.
    metadata
        metadata associated with the type, e.g. information about categorials.
    """
    if pa.types.is_signed_integer(t_pa):
        return pa.int64(), "int64", "int64", None
    elif pa.types.is_unsigned_integer(t_pa):
        return pa.uint64(), "uint64", "uint64", None
    elif pa.types.is_floating(t_pa):
        return pa.float64(), "float64", "float64", None
    elif pa.types.is_list(t_pa):
        assert t_pd is not None
        t_pa2, t_pd2, t_np2, metadata2 = normalize_type(
            t_pa.value_type, t_pd[len("list[") : -1], None, None
        )
        return pa.list_(t_pa2), f"list[{t_pd2}]", "object", None
    elif pa.types.is_dictionary(t_pa):
        return normalize_type(t_pa.value_type, t_pd, t_np, None)
    elif pa.types.is_string(t_pa) or pa.types.is_large_string(t_pa):
        # Pyarrow only supports reading back
        #
        # pyarrow + np.nan
        # pa.large_string(), "object", "str", None
        # or
        # python + pd.NA
        # pa.string(), "unicode", "string", None
        #
        # unintuitively, the numpy type identifier `t_np` corresponds
        # to the pandas dtypes `str` and `string`

        # pandas also supports mixed types but those are rare and must be
        # constructed explicitly
        if t_pd == "categorical":
            # We loose the information of the nullable type since the t_np type
            # is set to the dtype of the codes but not the categories.
            return pa.large_string(), "object", "str", None
        elif t_np == "str":
            return pa.large_string(), "object", "str", None
        elif t_np == "string":
            return pa.string(), "unicode", "string", None
        else:
            # This should be the ordinary object dtype
            return t_pa, t_pd, t_np, metadata
    else:
        return t_pa, t_pd, t_np, metadata


def _get_common_metadata_key(dataset_uuid, table):
    return f"{dataset_uuid}/{table}/{naming.TABLE_METADATA_FILE}"


def read_schema_metadata(
    dataset_uuid: str, store: KeyValueStore, table: str = SINGLE_TABLE
) -> SchemaWrapper:
    """Read schema and metadata from store.

    Parameters
    ----------
    dataset_uuid
        Unique ID of the dataset in question.
    store
        Object that implements `.get(key)` to read data.
    table
        Table to read metadata for.

    Returns
    -------
    schema: Schema
        Schema information for DataFrame/table.
    """
    key = _get_common_metadata_key(dataset_uuid=dataset_uuid, table=table)
    return SchemaWrapper(_bytes2schema(store.get(key)), key)


def store_schema_metadata(
    schema: SchemaWrapper,
    dataset_uuid: str,
    store: KeyValueStore,
    table: str = SINGLE_TABLE,
) -> str:
    """Store schema and metadata to store.

    Parameters
    ----------
    schema
        Schema information for DataFrame/table.
    dataset_uuid
        Unique ID of the dataset in question.
    store
        Object that implements `.put(key, data)` to write data.
    table
        Table to write metadata for.

    Returns
    -------
    key: str
        Key to which the metadata was written to.
    """
    key = _get_common_metadata_key(dataset_uuid=dataset_uuid, table=table)
    return store.put(key, _schema2bytes(schema.internal()))


def _schema2bytes(schema: SchemaWrapper) -> bytes:
    buf = pa.BufferOutputStream()
    pq.write_metadata(schema, buf, version=PARQUET_VERSION, coerce_timestamps="us")
    return buf.getvalue().to_pybytes()


def _bytes2schema(data: bytes) -> SchemaWrapper:
    reader = pa.BufferReader(data)
    schema = pq.read_schema(reader)
    fields = []
    for idx in range(len(schema)):
        f = schema[idx]

        # schema data recovered from parquet always contains timestamp data in us-granularity, but pandas will use
        # ns-granularity, so we re-align the two different worlds here
        if f.type == pa.timestamp("us"):
            f = pa.field(f.name, pa.timestamp("ns"))

        fields.append(f)
    return pa.schema(fields, schema.metadata)


def _pandas_in_schemas(schemas):
    """Check if any schema contains pandas metadata."""
    has_pandas = False
    for schema in schemas:
        if schema.metadata and b"pandas" in schema.metadata:
            has_pandas = True
    return has_pandas


def _determine_schemas_to_compare(
    schemas: Sequence[SchemaWrapper], ignore_pandas: bool
) -> tuple[SchemaWrapper | None, list[tuple[SchemaWrapper, list[str]]]]:
    """Iterate over a list of `pyarrow.Schema` objects and prepares them for
    comparison by picking a reference and determining all null columns.

    .. note::

        If pandas metadata exists, the version stored in the metadata is overwritten with the currently
        installed version since we expect to stay backwards compatible

    Returns
    -------
    reference: Schema
        A reference schema which is picked from the input list. The reference schema is guaranteed
        to be a schema having the least number of null columns of all input columns. The set of null
        columns is guaranteed to be a true subset of all null columns of all input schemas. If no such
        schema can be found, an Exception is raised
    list_of_schemas: List[Tuple[Schema, List]]
        A list holding pairs of (Schema, null_columns) where the null_columns are all columns which are null and
        must be removed before comparing the schemas
    """
    has_pandas = _pandas_in_schemas(schemas) and not ignore_pandas
    schemas_to_evaluate: list[tuple[SchemaWrapper, list[str]]] = []
    reference = None
    null_cols_in_reference = set()

    for schema in set(schemas):
        if not isinstance(schema, SchemaWrapper):
            schema = SchemaWrapper(schema, "__unknown__")

        if has_pandas:
            metadata = schema.metadata
            if metadata is None or b"pandas" not in metadata:
                raise ValueError(
                    "Pandas and non-Pandas schemas are not comparable. "
                    "Use ignore_pandas=True if you only want to compare "
                    "on Arrow level."
                )
            pandas_metadata = load_json(metadata[b"pandas"].decode("utf8"))

            # we don't care about the pandas version, since we assume it's safe
            # to read datasets that were written by older or newer versions.
            pandas_metadata["pandas_version"] = f"{pd.__version__}"

            metadata_clean = deepcopy(metadata)
            metadata_clean[b"pandas"] = _dict_to_binary(pandas_metadata)
            current = SchemaWrapper(pa.schema(schema, metadata_clean), schema.origin)
        else:
            current = schema

        # If a field is null we cannot compare it and must therefore reject it
        null_columns = {field.name for field in current if field.type == pa.null()}  # type: ignore

        # Determine a valid reference schema. A valid reference schema is considered to be the schema
        # of all input schemas with the least empty columns.
        # The reference schema ought to be a schema whose empty columns are a true subset for all sets
        # of empty columns. This ensures that the actual reference schema is the schema with the most
        # information possible. A schema which doesn't fulfil this requirement would weaken the
        # comparison and would allow for false positives

        # Trivial case
        if reference is None:
            reference = current
            null_cols_in_reference = null_columns
        # The reference has enough information to validate against current schema.
        # Append it to the list of schemas to be verified
        elif null_cols_in_reference.issubset(null_columns):
            schemas_to_evaluate.append((current, null_columns))
        # current schema includes all information of reference and more.
        # Add reference to schemas_to_evaluate and update reference
        elif null_columns.issubset(null_cols_in_reference):
            schemas_to_evaluate.append((reference, null_cols_in_reference))
            reference = current
            null_cols_in_reference = null_columns
        # If there is no clear subset available elect the schema with the least null columns as `reference`.
        # Iterate over the null columns of `reference` and replace it with a non-null field of the `current`
        # schema which recovers the loop invariant (null columns of `reference` is subset of `current`)
        else:
            if len(null_columns) < len(null_cols_in_reference):
                reference, current = current, reference
                null_cols_in_reference, null_columns = (
                    null_columns,
                    null_cols_in_reference,
                )

            for col in null_cols_in_reference - null_columns:
                # Enrich the information in the reference by grabbing the missing fields
                # from the current iteration. This assumes that we only check for global validity and
                # isn't relevant where the reference comes from.
                reference = _swap_fields_by_name(reference, current, col)
                null_cols_in_reference.remove(col)
            schemas_to_evaluate.append((current, null_columns))

    assert (reference is not None) or (not schemas_to_evaluate)

    return reference, schemas_to_evaluate


def _swap_fields_by_name(reference, current, field_name):
    current_field = current.field(field_name)
    reference_index = reference.get_field_index(field_name)
    return reference.set(reference_index, current_field)


def _strip_columns_from_schema(schema, field_names):
    stripped_schema = schema

    for name in field_names:
        ix = stripped_schema.get_field_index(name)
        if ix >= 0:
            stripped_schema = stripped_schema.remove(ix)
        else:
            # If the returned index is negative, the field doesn't exist in the schema.
            # This is most likely an indicator for incompatible schemas and we refuse to strip the schema
            # to not obfurscate the validation result
            _logger.warning(
                f"Unexpected field `{name}` encountered while trying to strip `null` columns.\n"
                f"Schema was:\n\n`{schema}`"
            )
            return schema
    return stripped_schema


def _remove_diff_header(diff):
    diff = list(diff)
    for ix, el in enumerate(diff):
        # This marks the first actual entry of the diff
        # e.g. @@ -1,5 + 2,5 @@
        if el.startswith("@"):
            return diff[ix:]
    return diff


def _diff_schemas(first, second):
    # see https://issues.apache.org/jira/browse/ARROW-4176

    first_pyarrow_info = str(first.remove_metadata())
    second_pyarrow_info = str(second.remove_metadata())
    pyarrow_diff = _remove_diff_header(
        difflib.unified_diff(
            str(first_pyarrow_info).splitlines(), str(second_pyarrow_info).splitlines()
        )
    )

    first_pandas_info = first.pandas_metadata
    second_pandas_info = second.pandas_metadata
    pandas_meta_diff = _remove_diff_header(
        difflib.unified_diff(
            pprint.pformat(first_pandas_info).splitlines(),
            pprint.pformat(second_pandas_info).splitlines(),
        )
    )

    diff_string = (
        "Arrow schema:\n"
        + "\n".join(pyarrow_diff)
        + "\n\nPandas_metadata:\n"
        + "\n".join(pandas_meta_diff)
    )

    return diff_string


def validate_compatible(schemas, ignore_pandas=False):
    """Validate that all schemas in a given list are compatible.

    Apart from the pandas version preserved in the schema metadata, schemas must be completely identical. That includes
    a perfect match of the whole metadata (except the pandas version) and pyarrow types.

    Use :meth:`make_meta` and :meth:`normalize_column_order` for type and column order normalization.

    In the case that all schemas don't contain any pandas metadata, we will check the Arrow
    schemas directly for compatibility.

    Parameters
    ----------
    schemas: List[Schema]
        Schema information from multiple sources, e.g. multiple partitions. List may be empty.
    ignore_pandas: bool
        Ignore the schema information given by Pandas an always use the Arrow schema.

    Returns
    -------
    schema: SchemaWrapper
        The reference schema which was tested against

    Raises
    ------
    ValueError
        At least two schemas are incompatible.
    """
    reference, schemas_to_evaluate = _determine_schemas_to_compare(
        schemas, ignore_pandas
    )

    for current, null_columns in schemas_to_evaluate:
        # We have schemas so the reference schema should be non-none.
        assert reference is not None
        # Compare each schema to the reference but ignore the null_cols and the Pandas schema information.
        reference_to_compare = _strip_columns_from_schema(
            reference, null_columns
        ).remove_metadata()
        current_to_compare = _strip_columns_from_schema(
            current, null_columns
        ).remove_metadata()

        def _fmt_origin(origin):
            origin = sorted(origin)
            # dask cuts of exception messages at 1k chars:
            #   https://github.com/dask/distributed/blob/6e0c0a6b90b1d3c/distributed/core.py#L964
            # therefore, we cut the the maximum length
            max_len = 200
            inner_msg = ", ".join(origin)
            ellipsis = "..."
            if len(inner_msg) > max_len + len(ellipsis):
                inner_msg = inner_msg[:max_len] + ellipsis
            return f"{{{inner_msg}}}"

        if reference_to_compare != current_to_compare:
            schema_diff = _diff_schemas(reference, current)
            exception_message = f"""Schema violation

Origin schema: {_fmt_origin(current.origin)}
Origin reference: {_fmt_origin(reference.origin)}

Diff:
{schema_diff}

Reference schema:
{str(reference)}"""
            raise ValueError(exception_message)

    # add all origins to result AFTER error checking, otherwise the error message would be pretty misleading due to the
    # reference containing all origins.
    if reference is None:
        return None
    else:
        return reference.with_origin(
            reduce(
                set.union,
                (schema.origin for schema, _null_columns in schemas_to_evaluate),
                reference.origin,
            )
        )


def validate_shared_columns(schemas, ignore_pandas=False):
    """Validate that columns that are shared amongst schemas are compatible.

    Only DataFrame columns are taken into account, other fields (like index data) are ignored. The following data must
    be an exact match:

    - metadata (as stored in the ``"columns"`` list of the ``b'pandas'`` schema metadata)
    - pyarrow type (that means that e.g. ``int8`` and ``int64`` are NOT compatible)

    Columns that are only present in a subset of the provided schemas must only be compatible for that subset, i.e.
    non-existing columns are ignored. The order of the columns in the provided schemas is irrelevant.

    Type normalization should be handled by :meth:`make_meta`.

    In the case that all schemas don't contain any pandas metadata, we will check the Arrow
    schemas directly for compatibility. Then the metadata information will not be checked
    (as it is non-existent).

    Parameters
    ----------
    schemas: List[Schema]
        Schema information from multiple sources, e.g. multiple tables. List may be empty.
    ignore_pandas: bool
        Ignore the schema information given by Pandas an always use the Arrow schema.

    Raises
    ------
    ValueError
        Incompatible columns were found.
    """
    seen: dict[str, Any] = {}
    has_pandas = _pandas_in_schemas(schemas) and not ignore_pandas

    for schema in schemas:
        if has_pandas:
            metadata = schema.metadata
            if metadata is None or b"pandas" not in metadata:
                raise ValueError(
                    "Pandas and non-Pandas schemas are not comparable. "
                    "Use ignore_pandas=True if you only want to compare "
                    "on Arrow level."
                )
            pandas_metadata = load_json(metadata[b"pandas"].decode("utf8"))

            columns = []
            for cmd in pandas_metadata["columns"]:
                name = cmd.get("name")
                if name is None:
                    continue
                columns.append(cmd["field_name"])
        else:
            columns = schema.names

        for col in columns:
            field_idx = schema.get_field_index(col)
            field = schema[field_idx]
            obj = (field, col)
            if col in seen:
                ref = seen[col]
                if pa.types.is_null(ref[0].type) or pa.types.is_null(field.type):
                    continue
                if ref != obj:
                    raise ValueError(
                        f'Found incompatible entries for column "{col}"\n{ref}\n{obj}'
                    )
            else:
                seen[col] = obj


def _dict_to_binary(dct):
    return simplejson.dumps(dct, sort_keys=True).encode("utf8")


def empty_dataframe_from_schema(
    schema, columns=None, date_as_object=False, coerce_temporal_nanoseconds=True
):
    """Create an empty DataFrame from provided schema.

    Parameters
    ----------
    schema: Schema
        Schema information of the new empty DataFrame.
    columns: Union[None, List[str]]
        Optional list of columns that should be part of the resulting DataFrame. All columns in that list must also be
        part of the provided schema.
    date_as_object: bool
        Cast dates to objects.
    coerce_temporal_nanoseconds: bool
        Coerce date32, date64, duration and timestamp units to nanoseconds to retain behaviour of pandas 1.x.
        Only applicable to pandas version >= 2.0 and PyArrow version >= 13.0.0.

    Returns
    -------
    DataFrame
        Empty DataFrame with requested columns and types.
    """
    # HACK: Cast bytes to object in metadata until Pandas bug is fixed: https://github.com/pandas-dev/pandas/issues/50127
    schema = schema_metadata_bytes_to_object(schema.internal())

    _coerce = {"coerce_temporal_nanoseconds": coerce_temporal_nanoseconds}
    df = schema.empty_table().to_pandas(date_as_object=date_as_object, **_coerce)

    df.columns = df.columns.map(ensure_string_type)
    if columns is not None:
        df = df[columns]

    return df



from pyarrow import Schema


def _check_contains_null(val):
    if isinstance(val, bytes):
        for byte in val:
            if isinstance(byte, bytes):
                compare_to = chr(0)
            else:
                compare_to = 0
            if byte == compare_to:
                return True
    return False


def ensure_unicode_string_type(obj):
    """ensures obj is a of native string type:"""
    if isinstance(obj, bytes):
        return obj.decode("utf8")
    else:
        return str(obj)


def schema_metadata_bytes_to_object(schema: Schema) -> Schema:
    meta = schema.metadata[b"pandas"].decode().replace("bytes", "object").encode()
    return schema.with_metadata({b"pandas": meta})



import inspect
import textwrap
from collections import defaultdict

import pytest

from plateau.core.docs import _PARAMETER_MAPPING, default_docs
from plateau.io.dask.bag import (
    build_dataset_indices__bag,
    read_dataset_as_dataframe_bag,
    read_dataset_as_metapartitions_bag,
    store_bag_as_dataset,
)
from plateau.io.dask.dataframe import (
    read_dataset_as_ddf,
    store_dataset_from_ddf,
    update_dataset_from_ddf,
)
from plateau.io.dask.delayed import (
    delete_dataset__delayed,
    read_dataset_as_delayed,
    read_dataset_as_delayed_metapartitions,
    store_delayed_as_dataset,
    update_dataset_from_delayed,
)
from plateau.io.eager import (
    build_dataset_indices,
    commit_dataset,
    create_empty_dataset_header,
    delete_dataset,
    garbage_collect_dataset,
    read_dataset_as_dataframes,
    read_dataset_as_metapartitions,
    read_table,
    store_dataframes_as_dataset,
    update_dataset_from_dataframes,
    write_single_partition,
)
from plateau.io.iter import (
    read_dataset_as_dataframes__iterator,
    read_dataset_as_metapartitions__iterator,
    store_dataframes_as_dataset__iter,
    update_dataset_from_dataframes__iter,
)


@pytest.mark.parametrize(
    "function",
    [
        read_dataset_as_metapartitions_bag,
        read_dataset_as_dataframe_bag,
        store_bag_as_dataset,
        build_dataset_indices__bag,
        read_dataset_as_ddf,
        store_dataset_from_ddf,
        update_dataset_from_ddf,
        delete_dataset__delayed,
        read_dataset_as_delayed_metapartitions,
        read_dataset_as_delayed,
        update_dataset_from_delayed,
        store_delayed_as_dataset,
        delete_dataset,
        read_dataset_as_dataframes,
        read_dataset_as_metapartitions,
        read_table,
        commit_dataset,
        store_dataframes_as_dataset,
        create_empty_dataset_header,
        write_single_partition,
        update_dataset_from_dataframes,
        build_dataset_indices,
        garbage_collect_dataset,
        read_dataset_as_metapartitions__iterator,
        read_dataset_as_dataframes__iterator,
        update_dataset_from_dataframes__iter,
        store_dataframes_as_dataset__iter,
    ],
)
def test_docs(function):
    docstrings = function.__doc__
    arguments = inspect.signature(function).parameters
    valid_docs = defaultdict(set)
    for arg in arguments:
        valid = _PARAMETER_MAPPING.get(arg, "Parameters") in docstrings
        if not valid:
            indented = textwrap.indent(
                _PARAMETER_MAPPING.get(arg, "Parameters"), " " * 4, lambda line: True
            )
            valid = indented.lstrip() in docstrings
        valid_docs[valid].add(arg)

    assert valid_docs[True]
    if valid_docs[False]:
        raise AssertionError(
            f"Wrong or missing docstrings for parameters {valid_docs[False]}.\n\n{docstrings}"
        )

    sorted_arguments = sorted(arguments)
    args_in_docs = [argument in docstrings for argument in sorted_arguments]

    assert all(args_in_docs), [
        sorted_arguments[ix] for ix, val in enumerate(args_in_docs) if val is False
    ]


def test_docs_duplicity():
    # This test ensures that if a keyword argument has been listed out in the docs as well as under _PARAMETER_MAPPING
    # then the parser does not end up pick the one under _PARAMETER_MAPPING
    @default_docs
    def dummy_function(store):
        """This is a dummy_function.

        Parameters
        -----------
        store: str
            This is an argument
        """

    resultant_docstrings = str(dummy_function.__doc__)
    assert "This is a dummy_function" in resultant_docstrings
    assert "This is an argument" in resultant_docstrings
    assert "Factory function producing a KeyValueStore" not in resultant_docstrings
